/* Copyright (c) 2013 tonomu
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include <stdio.h>
#include <stdlib.h>
#include "mbt.h"

static int get_header(FILE *ifp, struct header_st *header);
static int init_gtrk(struct track_st *track, int total);
static int get_track(FILE *ifp, struct track_st *track);
static int get_twoc(FILE *ifp, unsigned int *value);
static int get_fourc(FILE *ifp, unsigned int *value);
static int init_gblk(struct block_st *block, unsigned int max);
static int get_time(FILE *ifp, int blk, struct block_st *block);
static int get_var(FILE *ifp ,long *value);
static int get_event(FILE *ifp, int blk, struct block_st *block);
static int get_sdat(const int *msg, unsigned char **adr);
static int get_ddat(FILE *ifp, const int *msg, unsigned char **adr);
static int get_vdat(FILE *ifp, const int *msg, unsigned char **adr);
static int get_meta(FILE *ifp, const int *msg, unsigned char **adr);
static int get_syse(FILE *ifp, const int *msg, unsigned char **adr);
static int get_chain(FILE *ifp, int *len);
static int get_len(const int *len, const int cnt);

int Get_Data(FILE *ifp, struct midi_st *midi)
{
  int status;
  int trk;

  status = get_header(ifp, &midi->header);
  if (status < 0) {
    (void) printf("error: smf header\n");
    return status;
  }

  midi->track = (struct track_st *) malloc(midi->header.total
                                           * sizeof(struct track_st));
  if (midi->track == NULL) { return E_G_TRACK_MALLOC; }

  status = init_gtrk(midi->track, midi->header.total);
  if (status < 0) { return status; }

  for (trk = 0; trk < midi->header.total; trk++) {
    status = get_track(ifp, &midi->track[trk]);
    if (status < 0) {
      (void) printf("error: track %d\n", trk);
      return status;
    }
  }

  return STATUS_SUCCESS;
}

static int get_header(FILE *ifp, struct header_st *header)
{
  int status;
  unsigned int data;
  const unsigned int mthd = 0x4d546864U;

  status = get_fourc(ifp, &data);
  if (status < 0) { return status; }
  if (data != mthd) { return E_G_MTHD_DATA; }

  status = get_fourc(ifp, &data);
  if (status < 0) { return status; }
  if (data != 6) { return E_G_HEADER_SIZE; }

  status = get_twoc(ifp, &data);
  if (status < 0) { return status; }
  if (data >> 1) { return E_G_SMF_FORMAT; }
  header->format = (int) data;

  status = get_twoc(ifp, &data);
  if (status < 0) { return status; }
  header->total = (int) data;

  status = get_twoc(ifp, &data);
  if (status < 0) { return status; }
  header->tpqn = (int) data;

  return STATUS_SUCCESS;
}

static int init_gtrk(struct track_st *track, int total)
{
  while (total) {
    total--;
    track->block = NULL;
    track++;
  }
  return STATUS_SUCCESS;
}

static int get_track(FILE *ifp, struct track_st *track)
{
  const unsigned int mtrk = 0x4d54726bU;
  unsigned int data;
  unsigned int size;
  unsigned int max;
  unsigned int cnt = 0;
  unsigned int before;
  int blk = 0;
  int status;

  struct block_st *rblock;

  status = get_fourc(ifp, &data);
  if (status < 0) { return status; }
  if (data != mtrk) { return E_G_MTRK_DATA; }

  status = get_fourc(ifp, &size);
  if (status < 0) { return status; }
  if (size == 1) { return E_G_TRACK_LEN; }
  max = (size >> 1) + 1;

  track->block = (struct block_st *) malloc(max * sizeof(struct block_st));
  if (track->block == NULL) { return E_G_BLOCK_MALLOC; }

  status = init_gblk(track->block, max);
  if (status < 0) { return status; }

  while (cnt < size) {
    before = cnt;
    status = get_time(ifp, blk, &(track->block[blk]));
    if (status < 0) { return status; }
    cnt += status;

    status = get_event(ifp, blk, &(track->block[blk]));
    if (status < 0) { return status; }
    cnt += status;
    if (before > cnt) { return E_G_TRACK_DATA; }

    blk++;
  }
  if (size != cnt) { return E_G_TRACK_SIZE; }

  rblock = (struct block_st *) realloc(track->block,
                                       ((unsigned int) blk + 1U)
                                       * sizeof(struct block_st));
  if (rblock == NULL) { return E_G_BLOCK_REALLOC; }
  track->block = rblock;

  return STATUS_SUCCESS;
}

static int get_twoc(FILE *ifp, unsigned int *value)
{
  int status;

  status = fgetc(ifp);
  if (status == EOF) { return E_G_GET_CHAR; }
  *value = (unsigned int) status;

  status = fgetc(ifp);
  if (status == EOF) { return E_G_GET_CHAR; }
  *value = (*value << 8) | ((unsigned int) status);

  return STATUS_SUCCESS;
}

static int get_fourc(FILE *ifp, unsigned int *value)
{
  int status;

  status = fgetc(ifp);
  if (status == EOF) { return E_G_GET_CHAR; }
  *value = (unsigned int) status;

  status = fgetc(ifp);
  if (status == EOF) { return E_G_GET_CHAR; }
  *value = (*value << 8) | (unsigned int) status;

  status = fgetc(ifp);
  if (status == EOF) { return E_G_GET_CHAR; }
  *value = (*value << 8) | (unsigned int) status;

  status = fgetc(ifp);
  if (status == EOF) { return E_G_GET_CHAR; }
  *value = (*value << 8) | (unsigned int) status;

  return STATUS_SUCCESS;
}

static int init_gblk(struct block_st *block, unsigned int max)
{
  while (max) {
    max--;
    block->time = -1;
    block->event = NULL;
    block++;
  }
  return STATUS_SUCCESS;
}

static int get_time(FILE *ifp, int blk, struct block_st *block)
{
  int cnt;

  cnt = get_var(ifp, &(block->time));
  if (cnt < 0) { return cnt; }

  if (blk > 0) {
    block->time += (block - 1)->time;
  }
  if (block->time < 0) { return E_G_TIME_VALUE; }

  return cnt;
}

static int get_var(FILE *ifp, long *value)
{
  int data;

  data = fgetc(ifp);
  if (data == EOF) { return E_G_GET_CHAR; }
  if ((data & 0x80) == 0) { *value = (long) data; return 1; }

  *value = ((long) data & 0x7F) << 7;
  data = fgetc(ifp);
  if (data == EOF) { return E_G_GET_CHAR; }
  if ((data & 0x80) == 0) { *value |= (long) data; return 2; }

  *value = (*value | ((long) data & 0x7F)) << 7;
  data = fgetc(ifp);
  if (data == EOF) { return E_G_GET_CHAR; }
  if ((data & 0x80) == 0) { *value |= (long) data; return 3; }

  *value = (*value | ((long) data & 0x7F)) << 7;
  data = fgetc(ifp);
  if (data == EOF) { return E_G_GET_CHAR; }
  if ((data & 0x80) == 0) { *value |= (long) data; return 4; }

  return E_G_LEN_FORMAT;
}

static int get_event(FILE *ifp, int blk, struct block_st *block)
{
  int msg[2];
  int type;
  int rng;
  int cnt;

  msg[0] = fgetc(ifp);
  if (msg[0] == EOF) { return E_G_GET_CHAR; }

  if (msg[0] & 0x80) {
    msg[1] = fgetc(ifp);
    if (msg[1] == EOF) { return E_G_GET_CHAR; }
    rng = 0;
  }
  else {
    if (blk == 0) { return E_G_SET_RNG; }
    msg[1] = msg[0];
    msg[0] = (int) (block - 1)->event[0];
    rng = 1;
  }

  type = msg[0] & 0xF0;

  if      (type == 0x90) { cnt = get_ddat(ifp, msg, &(block->event)); }
  else if (type == 0x80) { cnt = get_ddat(ifp, msg, &(block->event)); }
  else if (type == 0xB0) { cnt = get_ddat(ifp, msg, &(block->event)); }
  else if (type == 0xF0) { cnt = get_vdat(ifp, msg, &(block->event)); }
  else if (type == 0xE0) { cnt = get_ddat(ifp, msg, &(block->event)); }
  else if (type == 0xA0) { cnt = get_ddat(ifp, msg, &(block->event)); }
  else                   { cnt = get_sdat(     msg, &(block->event)); }

  if (cnt < 0) { return cnt; }

  return cnt - rng;
}

static int get_sdat(const int *msg, unsigned char **adr)
{
  *adr = (unsigned char *) malloc(2 * sizeof(unsigned char));
  if (*adr == NULL) { return E_G_EVENT_MALLOC; }

  (*adr)[0] = (unsigned char) msg[0];
  (*adr)[1] = (unsigned char) msg[1];

  return 2;
}

static int get_ddat(FILE *ifp, const int *msg, unsigned char **adr)
{
  int data;

  *adr = (unsigned char *) malloc(3 * sizeof(unsigned char));
  if (*adr == NULL) { return E_G_EVENT_MALLOC; }

  (*adr)[0] = (unsigned char) msg[0];
  (*adr)[1] = (unsigned char) msg[1];

  data = fgetc(ifp);
  if (data == EOF) { return E_G_GET_CHAR; }
  (*adr)[2] = (unsigned char) data;

  return 3;
}

static int get_vdat(FILE *ifp, const int *msg, unsigned char **adr)
{
  if (*msg == 0xFF) { return get_meta(ifp, msg, adr); }
  if (*msg == 0xF0) { return get_syse(ifp, msg, adr); }
  if (*msg == 0xF7) { return get_syse(ifp, msg, adr); }

  return E_G_EVENT_TYPE;
}

static int get_meta(FILE *ifp, const int *msg, unsigned char **adr)
{
  int len[4];
  int cnt;
  int size;
  int i;
  int data;

  len[0] = fgetc(ifp);
  if (len[0] == EOF) { return E_G_GET_CHAR; }

  cnt = get_chain(ifp, len);
  if (cnt < 0) { return cnt; }

  size = get_len(len, cnt);
  if (size < 0) { return size; }

  size = 2 + cnt + size;
  *adr = (unsigned char *) malloc(size * sizeof(unsigned char));
  if (*adr == NULL) { return E_G_EVENT_MALLOC; }

  (*adr)[0] = (unsigned char) msg[0];
  (*adr)[1] = (unsigned char) msg[1];
  (*adr)[2] = (unsigned char) len[0];
  for (i = 1; i < cnt; i++) {
    (*adr)[2 + i] = (unsigned char) len[i];
  }

  for (i = 2 + cnt; i < size; i++) {
    data = fgetc(ifp);
    if (data == EOF) { return E_G_GET_CHAR; }
    (*adr)[i] = (unsigned char) data;
  }

  return size;
}

static int get_syse(FILE *ifp, const int *msg, unsigned char **adr)
{
  int len[4];
  int cnt;
  int size;
  int i;
  int data;

  len[0] = msg[1];

  cnt = get_chain(ifp, len);
  if (cnt < 0) { return cnt; }

  size = get_len(len, cnt);
  if (size < 0) { return size; }

  size = 1 + cnt + size;
  *adr = (unsigned char *) malloc(size * sizeof(unsigned char));
  if (*adr == NULL) { return E_G_EVENT_MALLOC; }

  (*adr)[0] = (unsigned char) msg[0];
  (*adr)[1] = (unsigned char) len[0];
  for (i = 1; i < cnt; i++) {
    (*adr)[1 + i] = (unsigned char) len[i];
  }

  for (i = 1 + cnt; i < size; i++) {
    data = fgetc(ifp);
    if (data == EOF) { return E_G_GET_CHAR; }
    (*adr)[i] = (unsigned char) data;
  }

  return size;
}

static int get_chain(FILE *ifp, int *len)
{
  if ((len[0] & 0x80) == 0) { return 1; }

  len[1] = fgetc(ifp);
  if (len[1] == EOF) { return E_G_GET_CHAR; }
  if ((len[1] & 0x80) == 0) { return 2; }

  len[2] = fgetc(ifp);
  if (len[2] == EOF) { return E_G_GET_CHAR; }
  if ((len[2] & 0x80) == 0) { return 3; }

  len[3] = fgetc(ifp);
  if (len[3] == EOF) { return E_G_GET_CHAR; }
  if ((len[3] & 0x80) == 0) { return 4; }

  return E_G_LEN_FORMAT;
}

static int get_len(const int *len, const int cnt)
{
  int value;
  if (cnt == 1) { return len[0]; }

  value = (len[0] & 0x7F) << 7;
  if (cnt == 2) { return value | len[1]; }

  value = (value | (len[1] & 0x7F)) << 7;
  if (cnt == 3) { return value | len[2]; }

  value = (value | (len[2] & 0x7F)) << 7;
  if (cnt == 4) { return value | len[3]; }

  return E_G_LEN_FORMAT;
}
