/* Copyright (c) 2013 tonomu
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include "mbt.h"

static int check_opt(char *argv[], struct arg_st *opt);
static int init_opt(struct arg_st *opt);
static int arg_switch(char *argv[], struct arg_st *opt);
static int arg_file(char *arg, char *file);
static int arg_num(char *arg, int *num);
static int show_help(void);
static int show_version(void);
static int calc_gcd(int num, int den);
static int change_tpqn(struct midi_st *midi, int gcd, int outq);
static int calc_tick(long *tick, int mul, int div);
static int open_infile(char *infile, FILE **ifp);
static int open_outfile(struct arg_st *opt, FILE **ofp);
static int file_name(char *infile, char *outfile, const int compile);
static int free_midi(struct midi_st *midi);

enum {
  DEFAULT_INPUT_TPQN = 480
};

int main(int argc, char *argv[])
{
  FILE *ifp;
  FILE *ofp;
  struct midi_st midi;
  struct arg_st opt;
  int gcd;
  int status;

  if (argc <= 1) { return show_help(); }

  status = check_opt(argv, &opt);
  if (status < 0) { return Print_Error(status); }

  if (opt.help) { return show_help(); }
  if (opt.version) { return show_version(); }

  status = open_infile(opt.infile, &ifp);
  if (status < 0) {
    (void) printf("error: input file\n");
    return Print_Error(status);
  }

  midi.track = NULL;

  if (opt.compile == 0) {
    if (opt.inq != 0) {
      (void) fclose(ifp);
      (void) free_midi(&midi);
      return Print_Error(E_M_INPUT_TPQN);
    }
    status = Get_Data(ifp, &midi);
  }
  else {
    if (opt.inq != 0) { midi.header.tpqn = opt.inq; }
    else { midi.header.tpqn = DEFAULT_INPUT_TPQN; }

    status = Read_Data(ifp, &midi);
  }
  if (status < 0) {
    (void) fclose(ifp);
    (void) free_midi(&midi);
    return Print_Error(status);
  }
  status = fclose(ifp);
  if (status == EOF) {
    (void) free_midi(&midi);
    return Print_Error(E_M_CLOSE_INPUT);
  }

  if (opt.outq) {
    gcd = calc_gcd(midi.header.tpqn, opt.outq);
    if (gcd <= 0) {
      (void) free_midi(&midi);
      return Print_Error(E_M_GCD_VALUE);
    }
    status = change_tpqn(&midi, gcd, opt.outq);
    if (status < 0) {
      (void) free_midi(&midi);
      return Print_Error(status);
    }
    midi.header.tpqn = opt.outq;
  }

  status = open_outfile(&opt, &ofp);
  if (status < 0) {
    (void) printf("error: output file\n");
    (void) free_midi(&midi);
    return Print_Error(status);
  }


  if (opt.compile == 0) {
    status = Write_Data(&midi, ofp);
  }
  else {
    if (opt.format == 0) {
      if (midi.header.total != 1) {
        (void) fclose(ofp);
        (void) free_midi(&midi);
        return Print_Error(E_M_FORMAT_ZERO);
      }
      midi.header.format = 0;
    }
    else {
      midi.header.format = 1;
    }
    status = Put_Data(&midi, ofp);
  }
  if (status < 0) {
    (void) fclose(ofp);
    (void) free_midi(&midi);
    return Print_Error(status);
  }

  status = fclose(ofp);
  if (status == EOF) {
    (void) free_midi(&midi);
    return Print_Error(E_M_CLOSE_OUTPUT);
  }

  status = free_midi(&midi);
  if (status < 0) { return Print_Error(status); }

  return STATUS_SUCCESS;
}

static int check_opt(char *argv[], struct arg_st *opt)
{
  int status;
  int cnt;

  status = init_opt(opt);
  if (status < 0) { return status; }

  argv++;
  while (*argv != NULL) {
    if ((*argv)[0] == '-') {
      cnt = arg_switch(argv, opt);
      if (cnt < 0) { return cnt; }
    }
    else {
      status = arg_file(*argv, opt->infile);
      if (status < 0) {
        (void) printf("error: input file\n");
        return status;
      }
      cnt = 1;
    }

    argv = argv + cnt;
  }

  return STATUS_SUCCESS;
}

static int init_opt(struct arg_st *opt)
{
  opt->compile = 0;
  opt->infile[0] = '\0';
  opt->outfile[0] = '\0';
  opt->format = 1;
  opt->inq = 0;
  opt->outq = 0;
  opt->help = 0;
  opt->version = 0;

  return STATUS_SUCCESS;
}

static int arg_switch(char *argv[], struct arg_st *opt)
{
  char *flag;
  int status;

  flag = *argv;
  flag++;

  if (*flag == '\0') { return E_M_FLAG_FORMAT; }

  while (*flag != '\0') {
    switch (*flag) {
    case 'c': opt->compile = 1; break;
    case 'z': opt->format  = 0; break;
    case 'v': opt->version = 1; break;
    case 'h': opt->help    = 1; break;
    case 'o':
      if (*(flag + 1) != '\0') { return E_M_ARGUMENT_END; }
      status = arg_file(*(argv + 1), opt->outfile);
      if (status < 0) {
        (void) printf("error: output file\n");
        return status;
      }
      return 2;
    case 't':
      if (*(flag + 1) != '\0') { return E_M_ARGUMENT_END; }
      status = arg_num(*(argv + 1), &(opt->outq));
      if (status < 0) {
        (void) printf("error: output tpqn\n");
        return status;
      }
      return 2;
    case 'q':
      if (*(flag + 1) != '\0') { return E_M_ARGUMENT_END; }
      status = arg_num(*(argv + 1), &(opt->inq));
      if (status < 0) {
        (void) printf("error: input tpqn\n");
        return status;
      }
      return 2;
    default: return E_M_FLAG_TYPE;
    }

    flag++;
  }
  return 1;
}

static int arg_file(char *arg, char *file)
{
  int i;

  if (arg == NULL) { return E_M_ASSIGN_FNAME; }

  if (file[0] != '\0') { return E_M_SET_FNAME; }

  for (i = 0; i < FILENAME_LEN - 1; i++) {
    file[i] = arg[i];
    if (arg[i] == '\0') { return STATUS_SUCCESS; }
  }
  file[i] = '\0';

  return E_M_FNAME_LEN;
}

static int arg_num(char *arg, int *num)
{

  if (arg == NULL) { return E_M_ASSIGN_DATA; }

  if (*num != 0) { return E_M_SET_TPQN; }

  while (*arg != '\0') {
    if (isdigit(*arg)) {
      *num = *num * 10;
      *num = *num + (int) (*arg - '0');
      if (*num >= 0x8000) { return E_M_TPQN_VALUE; }
    }
    else { return E_M_TPQN_DIGIT; }

    arg++;
  }

  if (*num == 0) { return E_M_TPQN_ZERO; }


  return STATUS_SUCCESS;
}

static int show_help(void)
{

  (void) printf("usage: mbt [options] file\n");
  (void) printf("options:\n");
  (void) printf(" -c           compile mode\n");
  (void) printf(" -h           display this information\n");
  (void) printf(" -v           display version\n");
  (void) printf(" -z           output SMF format 0 (when compile mode)\n");
  (void) printf(" -o <file>    set output <file>\n");
  (void) printf(" -q <TPQN>    set input <TPQN> (when compile mode)\n");
  (void) printf(" -t <TPQN>    set output <TPQN>\n");

  return STATUS_SUCCESS;
}

static int show_version(void)
{
  (void) printf("mbt version %d.%d.%d\n",
                MAJOR_VERSION,
                MINOR_VERSION,
                MODIFY_VERSION);

  return STATUS_SUCCESS;
}

static int calc_gcd(int num, int den)
{
  int rem;
  while (den != 0) {
    rem = num % den;
    num = den;
    den = rem;
  }
  return num;
}

static int change_tpqn(struct midi_st *midi, int gcd, int outq)
{
  struct track_st *track;
  struct block_st *block;
  int total;
  int div = midi->header.tpqn / gcd;
  int mul = outq / gcd;
  int status;

  total = midi->header.total;
  track = midi->track;

  while (total) {
    block = track->block;
    while (block->event != NULL) {
      status = calc_tick(&(block->time), mul, div);
      if (status < 0) { return status; }
      block++;
    }
    total--;
    track++;
  }
  return STATUS_SUCCESS;
}

static int calc_tick(long *tick, int mul, int div)
{
  long up;
  long lw;
  long over;
  const long hsize = sizeof(long) << 2;
  const long hmask = ~0UL >> (sizeof(long) << 2);

  up = *tick >> hsize;
  lw = *tick & hmask;

  up = up * mul;
  lw = lw * mul;

  over = (up + (lw >> hsize)) >> (hsize - 1);
  if (over) {
    *tick = *tick / div;
    up = *tick >> hsize;
    lw = *tick & hmask;

    up = up * mul;
    lw = lw * mul;

    over = (up + (lw >> hsize)) >> (hsize - 1);
    if (over) { return E_M_TICK_VALUE; }

    *tick = (up << hsize) + lw;
    return STATUS_SUCCESS;
  }

  *tick = ((up << hsize) + lw) / div;

  return STATUS_SUCCESS;

}

static int open_infile(char *infile, FILE **ifp)
{
  if (*infile == '\0') { return E_M_INPUT_FNAME; }

  *ifp = fopen(infile, "rb");
  if (*ifp == NULL) { return E_M_OPEN_INPUT; }

  return STATUS_SUCCESS;
}

static int open_outfile(struct arg_st *opt, FILE **ofp)
{
  if (*(opt->outfile) == '\0') {
    file_name(opt->infile, opt->outfile, opt->compile);
  }

  if (opt->compile == 0) {
    *ofp = fopen(opt->outfile, "w");
  }
  else {
    *ofp = fopen(opt->outfile, "wb");
  }

  if (*ofp == NULL) { return E_M_OPEN_OUTPUT; }

  return STATUS_SUCCESS;
}

static int file_name(char *infile, char *outfile, const int compile)
{
  int i = 0;
  int period = 0;

  while (infile[i] != '\0') {
    if (infile[i] == '.') { period = i; }
    outfile[i] = infile[i];
    i++;
  }

  if (period > FILENAME_LEN - 5) { return E_M_FNAME_LEN; }
  if (period) { i = period; }

  if (compile) {
    outfile[i]     = '.';
    outfile[i + 1] = 'm';
    outfile[i + 2] = 'i';
    outfile[i + 3] = 'd';
    outfile[i + 4] = '\0';
    return STATUS_SUCCESS;
  }
  outfile[i]     = '.';
  outfile[i + 1] = 'm';
  outfile[i + 2] = 'b';
  outfile[i + 3] = 't';
  outfile[i + 4] = '\0';
  return STATUS_SUCCESS;
}

static int free_midi(struct midi_st *midi)
{
  int i;
  int j;

  if (midi->track == NULL) { return STATUS_SUCCESS; }

  for (i = 0; i < midi->header.total; i++) {
    if (midi->track[i].block != NULL) {
      j = 0;
      while (midi->track[i].block[j].event != NULL) {
        free(midi->track[i].block[j].event);
        j++;
      }
      free(midi->track[i].block);
    }
  }
  free(midi->track);

  return STATUS_SUCCESS;
}
