/* Copyright (c) 2013 tonomu
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include <stdio.h>
#include "mbt.h"

static int put_header(struct midi_st *midi, FILE *ofp);
static int put_mthd(FILE *ofp);
static int put_six(FILE *ofp);
static int put_format(const int format, FILE *ofp);
static int put_total(const int total, FILE *ofp);
static int put_tpqn(const int tpqn, FILE *ofp);
static int put_track(struct block_st *block, FILE *ofp);
static int put_mtrk(FILE *ofp);
static int put_zero(FILE *ofp);
static int put_block(struct block_st *block, unsigned int *size, FILE *ofp);
static int put_size(const unsigned int size, FILE *ofp);
static int put_delta(const long time, FILE *ofp);
static int put_event(const unsigned char *event, FILE *ofp);
static int put_sdat(const unsigned char *event, FILE *ofp);
static int put_ddat(const unsigned char *event, FILE *ofp);
static int put_vdat(const unsigned char *event, FILE *ofp);
static int put_syse(const unsigned char *event, FILE *ofp);
static int put_meta(const unsigned char *event, FILE *ofp);
static int put_chain(const unsigned char *dat);
static int put_len(const unsigned char *len, const int cnt);

int Put_Data(struct midi_st *midi, FILE *ofp)
{
  int i;
  int status;

  if (midi->header.total > 0xFFFF) { return E_P_TRACK_COUNT; }
  if (midi->header.tpqn >= 0x8000) { return E_P_TPQN_VALUE; }

  status = put_header(midi, ofp);
  if (status < 0) { return status; }

  for (i = 0; i < midi->header.total; i++) {
    status = put_track(midi->track[i].block, ofp);
    if (status < 0) {
      (void) printf("error: track %d\n", i);
      return status;
    }
  }

  return STATUS_SUCCESS;
}

static int put_header(struct midi_st *midi, FILE *ofp)
{
  int status;

  status = put_mthd(ofp);
  if (status < 0) { return status; }

  status = put_six(ofp);
  if (status < 0) { return status; }

  status = put_format(midi->header.format, ofp);
  if (status < 0) { return status; }

  status = put_total(midi->header.total, ofp);
  if (status < 0) { return status; }

  status = put_tpqn(midi->header.tpqn, ofp);
  if (status < 0) { return status; }

  return STATUS_SUCCESS;
}

static int put_mthd(FILE *ofp)
{
  int status;

  status = fputs("MThd", ofp);
  if (status == EOF) { return E_P_OUT_MTHD; }

  return STATUS_SUCCESS;
}

static int put_six(FILE *ofp)
{
  int status;

  status = fputc(0x00, ofp);
  if (status == EOF) { return E_P_OUT_HEADSIZE; }

  status = fputc(0x00, ofp);
  if (status == EOF) { return E_P_OUT_HEADSIZE; }

  status = fputc(0x00, ofp);
  if (status == EOF) { return E_P_OUT_HEADSIZE; }

  status = fputc(0x06, ofp);
  if (status == EOF) { return E_P_OUT_HEADSIZE; }

  return STATUS_SUCCESS;
}

static int put_format(const int format, FILE *ofp)
{
  int status;

  if (format >> 1) { return E_P_FORMAT_TYPE; }

  status = fputc(0x00, ofp);
  if (status == EOF) { return E_P_OUT_FORMAT; }

  status = fputc((format & 0x01), ofp);
  if (status == EOF) { return E_P_OUT_FORMAT; }

  return STATUS_SUCCESS;
}

static int put_total(const int total, FILE *ofp)
{
  int status;

  status = fputc(((total >> 8) & 0xFF), ofp);
  if (status == EOF) { return E_P_OUT_TRACK; }

  status = fputc((total & 0xFF), ofp);
  if (status == EOF) { return E_P_OUT_TRACK; }

  return STATUS_SUCCESS;
}

static int put_tpqn(const int tpqn, FILE *ofp)
{
  int status;

  status = fputc(((tpqn >> 8) & 0xFF), ofp);
  if (status == EOF) { return E_P_OUT_TPQN; }

  status = fputc((tpqn & 0xFF), ofp);
  if (status == EOF) { return E_P_OUT_TPQN; }

  return STATUS_SUCCESS;
}

static int put_track(struct block_st *block, FILE *ofp)
{
  unsigned int size;
  int status;
  fpos_t start;
  fpos_t end;

  status = put_mtrk(ofp);
  if (status < 0) { return status; }

  status = fgetpos(ofp, &start);
  if (status != 0) { return E_P_POP_START; }

  status = put_zero(ofp);
  if (status < 0) { return status; }

  status = put_block(block, &size, ofp);
  if (status < 0) { return status; }

  status = fgetpos(ofp, &end);
  if (status != 0) { return E_P_POP_END; }

  status = fsetpos(ofp, &start);
  if (status != 0) { return E_P_SET_START; }

  status = put_size(size, ofp);
  if (status < 0) { return status; }

  status = fsetpos(ofp, &end);
  if (status != 0) { return E_P_SET_END; }

  return STATUS_SUCCESS;
}

static int put_mtrk(FILE *ofp)
{
  int status;

  status = fputs("MTrk", ofp);
  if (status == EOF) { return E_P_OUT_MTRK; }

  return STATUS_SUCCESS;
}

static int put_zero(FILE *ofp)
{
  int status;

  status = fputc(0x00, ofp);
  if (status == EOF) { return E_P_OUT_TEMPSIZE; }
  status = fputc(0x00, ofp);
  if (status == EOF) { return E_P_OUT_TEMPSIZE; }
  status = fputc(0x00, ofp);
  if (status == EOF) { return E_P_OUT_TEMPSIZE; }
  status = fputc(0x00, ofp);
  if (status == EOF) { return E_P_OUT_TEMPSIZE; }

  return STATUS_SUCCESS;
}

static int put_block(struct block_st *block, unsigned int *size, FILE *ofp)
{
  unsigned int before;
  int cnt;
  long delta = 0;

  *size = 0;

  while (block->event != NULL) {
    cnt = put_delta(block->time - delta, ofp);
    if (cnt <= 0) { return cnt; }

    before = *size;
    *size += cnt;
    if (*size < before) { return E_P_TRACK_SIZE; }

    delta = block->time;

    cnt = put_event(block->event, ofp);
    if (cnt <= 0) { return cnt; }

    before = *size;
    *size += cnt;
    if (*size < before) { return E_P_TRACK_SIZE; }

    block++;
  }

  return STATUS_SUCCESS;

}

static int put_size(const unsigned int size, FILE *ofp)
{
  int status;

  status = fputc((int) ((size >> 24) & 0xFF), ofp);
  if (status == EOF) { return E_P_OUT_DATASIZE; }

  status = fputc((int) ((size >> 16) & 0xFF), ofp);
  if (status == EOF) { return E_P_OUT_DATASIZE; }

  status = fputc((int) ((size >> 8) & 0xFF), ofp);
  if (status == EOF) { return E_P_OUT_DATASIZE; }

  status = fputc((int) (size & 0xFF), ofp);
  if (status == EOF) { return E_P_OUT_DATASIZE; }

  return STATUS_SUCCESS;
}

static int put_delta(const long time, FILE *ofp)
{
  int status;

  if (time < 0) { return E_P_DELTA_TICK; }

  if (time < 0x80) {
    status = fputc((int) time, ofp);
    if (status == EOF) { return E_P_OUT_TICK; }
    return 1;
  }
  if (time < 0x4000) {
    status = fputc((int) (((time >> 7) & 0x7F) | 0x80), ofp);
    if (status == EOF) { return E_P_OUT_TICK; }
    status = fputc((int) (time & 0x7F), ofp);
    if (status == EOF) { return E_P_OUT_TICK; }
    return 2;
  }
  if (time < 0x200000) {
    status = fputc((int) (((time >> 14) & 0x7F) | 0x80), ofp);
    if (status == EOF) { return E_P_OUT_TICK; }
    status = fputc((int) (((time >> 7) & 0x7F) | 0x80), ofp);
    if (status == EOF) { return E_P_OUT_TICK; }
    status = fputc((int) (time & 0x7F), ofp);
    if (status == EOF) { return E_P_OUT_TICK; }
    return 3;
  }
  if (time < 0x10000000) {
    status = fputc((int) (((time >> 21) & 0x7F) | 0x80), ofp);
    if (status == EOF) { return E_P_OUT_TICK; }
    status = fputc((int) (((time >> 14) & 0x7F) | 0x80), ofp);
    if (status == EOF) { return E_P_OUT_TICK; }
    status = fputc((int) (((time >> 7) & 0x7F) | 0x80), ofp);
    if (status == EOF) { return E_P_OUT_TICK; }
    status = fputc((int) (time & 0x7F), ofp);
    if (status == EOF) { return E_P_OUT_TICK; }
    return 4;
  }

  return E_P_DELTA_TICK;
}

static int put_event(const unsigned char *event, FILE *ofp)
{
  int type;

  type = (int) (event[0] & 0xF0);
  if (type == 0x90) { return put_ddat(event, ofp); }
  if (type == 0x80) { return put_ddat(event, ofp); }
  if (type == 0xB0) { return put_ddat(event, ofp); }
  if (type == 0xF0) { return put_vdat(event, ofp); }
  if (type == 0xE0) { return put_ddat(event, ofp); }
  if (type == 0xA0) { return put_ddat(event, ofp); }
  return put_sdat(event, ofp);
}

static int put_sdat(const unsigned char *event, FILE *ofp)
{
  int status;

  status = fputc((int) event[0], ofp);
  if (status == EOF) { return E_P_OUT_EVENTDATA; }
  status = fputc((int) event[1], ofp);
  if (status == EOF) { return E_P_OUT_EVENTDATA; }

  return 2;
}

static int put_ddat(const unsigned char *event, FILE *ofp)
{
  int status;

  status = fputc((int) event[0], ofp);
  if (status == EOF) { return E_P_OUT_EVENTDATA; }
  status = fputc((int) event[1], ofp);
  if (status == EOF) { return E_P_OUT_EVENTDATA; }
  status = fputc((int) event[2], ofp);
  if (status == EOF) { return E_P_OUT_EVENTDATA; }

  return 3;
}

static int put_vdat(const unsigned char *event, FILE *ofp)
{
  if (*event == 0xFF) { return put_meta(event, ofp); }
  if (*event == 0xF0) { return put_syse(event, ofp); }
  if (*event == 0xF7) { return put_syse(event, ofp); }

  return E_P_EVENT_TYPE;
}

static int put_syse(const unsigned char *event, FILE *ofp)
{
  int status;
  int cnt;
  int len;

  cnt = put_chain(event + 1);
  if (cnt < 0) { return cnt; }

  len = put_len(event + 1, cnt);
  if (len < 0) { return len; }

  len = 1 + cnt + len;
  cnt = len;

  while (len) {
    status = fputc((int) *event, ofp);
    if (status == EOF) { return E_P_OUT_EVENTDATA; }
    event++;
    len--;
  }

  return cnt;
}

static int put_meta(const unsigned char *event, FILE *ofp)
{
  int status;
  int cnt;
  int len;

  cnt = put_chain(event + 2);
  if (cnt < 0) { return cnt; }

  len = put_len(event + 2, cnt);
  if (len < 0) { return len; }

  len = 2 + cnt + len;
  cnt = len;

  while (len) {
    status = fputc((int) *event, ofp);
    if (status == EOF) { return E_P_OUT_EVENTDATA; }
    event++;
    len--;
  }
  return cnt;
}

static int put_chain(const unsigned char *dat)
{
  if ((dat[0] & 0x80) == 0) { return 1; }
  if ((dat[1] & 0x80) == 0) { return 2; }
  if ((dat[2] & 0x80) == 0) { return 3; }
  if ((dat[3] & 0x80) == 0) { return 4; }
  return E_P_DATA_CHAIN;
}

static int put_len(const unsigned char *len, const int cnt)
{
  int value;
  if (cnt == 1) { return (int) len[0]; }

  value = (int) (len[0] & 0x7F) << 7;
  if (cnt == 2) { return value | (int) len[1]; }

  value = (value | ((int) (len[1] & 0x7F))) << 7;
  if (cnt == 3) { return value | (int) len[2]; }

  value = (value | ((int) (len[2] & 0x7F))) << 7;
  if (cnt == 4) { return value | (int) len[3]; }

  return E_P_DATA_LEN;
}
