/* Copyright (c) 2013 tonomu
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include "mbt.h"

static int init_data(struct midi_st *midi, struct rmbt_st *inf);
static int init_rtrk(struct track_st *track, int start, int end);
static int init_rblk(struct block_st *block, int start, int end);
static int init_mbt(long *mbt);
static int read_pos(struct midi_st *midi, struct rmbt_st *inf);
static int read_file(FILE *ifp, struct rmbt_st *inf);
static int read_eof(struct rmbt_st *inf);
static int read_ret(struct rmbt_st *inf);
static int read_crr(FILE *ifp, struct rmbt_st *inf);
static int read_scl(FILE *ifp, struct rmbt_st *inf);
static int read_col(struct rmbt_st *inf);
static int read_dqt(FILE *ifp, struct rmbt_st *inf);
static int store_buf(const unsigned char ch, struct revt_st *buf);
static int init_buf(struct revt_st *buf);
static int read_line(struct midi_st *midi, struct rmbt_st *inf);
static int realloc_block(struct rmbt_st *inf);
static int read_time(struct rmbt_st *inf);
static int read_label(unsigned char **adr, long *label);
static int read_delta(unsigned char **adr, long *label);
static int read_mbt(unsigned char **adr, long *label);
static int read_skip(unsigned char **adr);
static int make_mbt(struct rmbt_st *inf, const int cnt, const long *label);
static int make_bar(struct rmbt_st *inf, const long *label);
static int store_mbt(struct rmbt_st *inf);
static int read_event(struct revt_st *buf);
static int read_note(struct revt_st *buf);
static int read_sdat(struct revt_st *buf);
static int read_ddat(struct revt_st *buf);
static int read_fxmg(struct revt_st *buf);
static int read_syse(struct revt_st *buf);
static int read_meta(struct revt_st *buf);
static int read_chain(const unsigned char *data);
static int read_len(const unsigned char *len, const int cnt);
static int read_long(unsigned char **adr, long *number);
static int read_carry(long *number, const unsigned char digit);
static int read_char(struct revt_st *buf);
static int istone(const unsigned char tone);
static int read_hex(struct revt_st *buf);
static int read_upper(struct revt_st *buf);
static int read_lower(struct revt_st *buf);
static int read_value(struct revt_st *buf);
static int read_tone(struct revt_st *buf);
static int read_step(const unsigned char in);
static int read_pitch(struct revt_st *buf);
static int read_tempo(struct revt_st *buf);
static int read_text(struct revt_st *buf);
static int read_esc(unsigned char *in);
static int shift_text(const int cnt, unsigned char *str);
static int read_var(int cnt, unsigned char **adr);
static int read_minus(struct revt_st *buf);
static int read_den(struct revt_st *buf);
static int read_exp(struct revt_st *buf);
static int read_pgrs(struct revt_st *buf);
static int read_num(unsigned char **adr, int max);
static int search_char(unsigned char **adr);
static int isskip(const unsigned char c);
static int store_ch(struct revt_st *buf);
static int store_den(struct midi_st *midi, struct rmbt_st *inf);
static int store_event(struct rmbt_st *inf);
static int store_mute(struct rmbt_st *inf);
static int store_block(struct rmbt_st *inf);
static int sort_block(struct block_st *block, int blk);
static int store_eot(struct midi_st *midi, struct rmbt_st *inf);
static int free_inf(struct rmbt_st *inf);

enum {
  ADD_PREVIOUS_TIME = 4,
  NONE_DUR = -1,
  HEX_END = 1,
  TRACK_COUNT_STEP = 64,
  BLOCK_COUNT_STEP = 8192,
  LINE_COUNT_STEP = 1024,
  MEAS_COUNT_STEP = 512,
  FIND_CHAR = 1,
  FIND_NULL = 0,
  COLUMN_OFFSET = 1
};

int Read_Data(FILE *ifp, struct midi_st *midi)
{
  struct rmbt_st inf;
  int status;

  status = init_data(midi, &inf);
  if (status < 0) {
    (void) free_inf(&inf);
    return status;
  }

  while (inf.eof != 1) {
    status = read_file(ifp, &inf);
    if (status < 0) {
      (void) read_pos(midi, &inf);
      return status;
    }

    status = init_buf(&(inf.buf));
    if (status < 0) { return status; }

    if (inf.sep >= 0) {
      status = read_line(midi, &inf);
    }
    else {
      status = search_char(&(inf.buf.in));
      if (status != FIND_NULL) { status = E_R_LINE_DATA; }
    }

    if (status < 0) {
      inf.buf.cnt = (long) (inf.buf.in - inf.buf.top);
      (void) read_pos(midi, &inf);
      return status;
    }
  }

  midi->header.total = inf.trk;

  status = free_inf(&inf);
  if (status < 0) { return status; }

  return STATUS_SUCCESS;
}

static int init_data(struct midi_st *midi, struct rmbt_st *inf)
{
  int status;

  midi->header.total = 0;
  inf->eof = 0;
  inf->row = 0;
  inf->blk = 0;
  inf->block = NULL;
  inf->buf.top = NULL;
  inf->bar = NULL;
  inf->trk = 0;

  if (midi->header.tpqn >= 0x8000) { return E_R_TPQN_VALUE; }
  if (midi->header.tpqn <= 0) { return E_R_TPQN_DATA; }

  inf->eot = TRACK_COUNT_STEP - 1;
  midi->track = (struct track_st *) malloc(inf->eot
                                           * sizeof(struct track_st));
  if (midi->track == NULL) { return E_R_TRACK_MALLOC; }

  status = init_rtrk(midi->track, 0, inf->eot);
  if (status < 0) { return status; }

  inf->eob = BLOCK_COUNT_STEP - 1;
  inf->block =
    (struct block_st *) malloc(inf->eob * sizeof(struct block_st));
  if (inf->block == NULL) { return E_R_INFBLOCK_MALLOC; }

  status = init_rblk(inf->block, 0, inf->eob);
  if (status < 0) { return status; }

  inf->buf.eol = LINE_COUNT_STEP - 1;
  inf->buf.top =
    (unsigned char *) malloc(inf->buf.eol * sizeof(unsigned char));
  if (inf->buf.top == NULL) { return E_R_INFBUF_MALLOC; }
  inf->buf.ch = 0xFF;

  inf->max = 0;
  inf->lim = MEAS_COUNT_STEP - 2;
  inf->bar =
    (struct rbar_st *) malloc(inf->lim * sizeof(struct rbar_st));
  if (inf->bar == NULL) { return E_R_INFBAR_MALLOC; }
  inf->bar[0].tick = 0;
  inf->bar[0].beat.num = 4;
  inf->bar[0].beat.den = midi->header.tpqn;

  status = init_mbt(inf->mbt);
  if (status < 0) { return status; }

  return STATUS_SUCCESS;
}

static int init_rtrk(struct track_st *track, int start, int end)
{
  track = track + start;
  end = end - start;
  while (end) {
    end--;
    track->block = NULL;
    track++;
  }
  return STATUS_SUCCESS;
}

static int init_rblk(struct block_st *block, int start, int end)
{
  block = block + start;
  end = end - start;
  while (end) {
    end--;
    block->time = -1;
    block->event = NULL;
    block++;
  }
  return STATUS_SUCCESS;
}

static int init_mbt(long *mbt)
{
  mbt[0] = 0;
  mbt[1] = 1;
  mbt[2] = 0;
  return STATUS_SUCCESS;
}

static int read_pos(struct midi_st *midi, struct rmbt_st *inf)
{
  midi->header.total = inf->trk;
  (void) printf("error: input file %ld, %ld\n",
                 inf->row, inf->buf.cnt + COLUMN_OFFSET);
  (void) printf("error: track %d\n", inf->trk);
  (void) free_inf(inf);

  return STATUS_SUCCESS;
}

static int read_file(FILE *ifp, struct rmbt_st *inf)
{
  int value;
  int status;

  inf->sep = -1;
  inf->buf.cnt = 0;
  (inf->row)++;
  if (inf->row <= 0) { return E_R_ROW_VALUE; }

  do {
    value = fgetc(ifp);
    if (value == EOF) {
      return read_eof(inf);
    }
    else if (value == '\n') {
      return read_ret(inf);
    }
    else if (value == '\r') {
      return read_crr(ifp, inf);
    }
    else if (value == ';') {
      return read_scl(ifp, inf);
    }
    else if (value == ':') {
      status = read_col(inf);
    }
    else if (value == '\"') {
      status = read_dqt(ifp, inf);
    }
    else {
      status = store_buf((unsigned char) (value & 0xFF), &inf->buf);
    }
  } while (status >= 0);

  return status;
}

static int read_eof(struct rmbt_st *inf)
{
  int value;
  value = store_buf('\0', &inf->buf);
  if (value < 0) { return value; }

  inf->eof = 1;
  return STATUS_SUCCESS;
}

static int read_ret(struct rmbt_st *inf)
{
  int value;
  value = store_buf('\0', &inf->buf);
  if (value < 0) { return value; }

  return STATUS_SUCCESS;
}

static int read_crr(FILE *ifp, struct rmbt_st *inf)
{
  int value;

  value = store_buf('\0', &inf->buf);
  if (value < 0) { return value; }

  value = fgetc(ifp);
  if (value == EOF) {
    inf->eof = 1;
    return STATUS_SUCCESS;
  }
  if (value != '\n') {
    value = ungetc(value, ifp);
    if (value == EOF) { return E_R_STACK_CHAR; }
  }

  return STATUS_SUCCESS;
}

static int read_scl(FILE *ifp, struct rmbt_st *inf)
{
  int value;
  value = store_buf('\0', &inf->buf);
  if (value < 0) { return value; }

  do {
    value = fgetc(ifp);
    if (value == EOF) {
      inf->eof = 1;
      break;
    }
    if (value == '\r') {
      value = fgetc(ifp);
      if (value == EOF) {
        inf->eof = 1;
        break;
      }
      if (value != '\n') {
        value = ungetc(value, ifp);
        if (value == EOF) { return E_R_STACK_CHAR; }
      }
      break;
    }
  } while (value != '\n');

  return STATUS_SUCCESS;
}

static int read_col(struct rmbt_st *inf)
{
  int status;

  if (inf->sep >= 0) { return E_R_SET_SEPARATOR; }

  status = store_buf('\0', &inf->buf);
  if (status < 0) { return status; }

  inf->sep = inf->buf.cnt;

  return STATUS_SUCCESS;
}

static int read_dqt(FILE *ifp, struct rmbt_st *inf)
{
  int value;
  int esc;
  int status;
  value = store_buf('\"', &inf->buf);
  if (value < 0) { return value; }

  do {
    value = fgetc(ifp);
    if (value == EOF)  { return E_R_EOF_POS; }
    if (value == '\n') { return E_R_RET_POS; }
    if (value == '\r') { return E_R_RET_POS; }

    status = store_buf((unsigned char) (value & 0xFF), &inf->buf);
    if (status < 0) { return status; }

    if (value == '\\') {
      esc = fgetc(ifp);
      if (esc == EOF)  { return E_R_EOF_POS; }
      if (esc == '\n') { return E_R_RET_POS; }
      if (esc == '\r') { return E_R_RET_POS; }

      status = store_buf((unsigned char) (esc & 0xFF), &inf->buf);
      if (status < 0) { return status; }
    }
  } while (value != '\"');

  return STATUS_SUCCESS;
}

static int store_buf(const unsigned char ch, struct revt_st *buf)
{
  unsigned char *rtop;

  if (buf->cnt >= buf->eol) {
    buf->eol += LINE_COUNT_STEP;
    if (buf->eol <= 0) { return E_R_LINE_VALUE; }

    rtop = (unsigned char *) realloc(buf->top,
                                     buf->eol * sizeof(unsigned char));
    if (rtop == NULL) { return E_R_INFBUF_REALLOC; }
    buf->top = rtop;
  }
  buf->top[buf->cnt] = ch;
  (buf->cnt)++;

  return STATUS_SUCCESS;
}

static int init_buf(struct revt_st *buf)
{
  buf->in  = buf->top;
  buf->out = buf->top;
  buf->cnt = 0;

  return STATUS_SUCCESS;
}

static int read_line(struct midi_st *midi, struct rmbt_st *inf)
{
  int status;

  status = realloc_block(inf);
  if (status < 0) { return status; }

  status = read_time(inf);
  if (status < 0) { return status; }

  status = store_mbt(inf);
  if (status < 0) { return status; }

  status = read_event(&(inf->buf));
  if (status < 0) { return status; }

  status = store_ch(&(inf->buf));
  if (status < 0) { return status; }

  status = store_den(midi, inf);
  if (status < 0) { return status; }

  status = store_event(inf);
  if (status < 0) { return status; }

  status = store_mute(inf);
  if (status < 0) { return status; }

  status = store_block(inf);
  if (status < 0) { return status; }

  status = store_eot(midi, inf);
  if (status < 0) { return status; }

  return STATUS_SUCCESS;
}

static int realloc_block(struct rmbt_st *inf)
{
  struct block_st *rblock;
  int status;

  if (inf->blk < inf->eob) { return STATUS_SUCCESS; }

  inf->eob += BLOCK_COUNT_STEP;
  if (inf->eob <= 0) {
    inf->eob = inf->blk;
    return E_R_BLOCK_COUNT;
  }

  rblock = (struct block_st *) realloc(inf->block,
                                       inf->eob * sizeof(struct block_st));
  if (rblock == NULL) {
    inf->eob = inf->blk;
    return E_R_INFBLOCK_REALLOC;
  }
  inf->block = rblock;

  status = init_rblk(inf->block, inf->blk, inf->eob);
  if (status < 0) { return status; }

  return STATUS_SUCCESS;
}

static int read_time(struct rmbt_st *inf)
{
  int count;
  long label[3];
  int status;

  count = read_label(&inf->buf.in, label);
  if (count < 0) { return count; }

  status = make_mbt(inf, count, label);
  if (status < 0) { return status; }

  return STATUS_SUCCESS;

}

static int read_label(unsigned char **adr, long *label)
{
  int status;

  status = read_skip(adr);
  if (status == FIND_NULL) { return 0; }

  if (**adr == '+') { return read_delta(adr, label); }
  return read_mbt(adr, label);
}

static int read_delta(unsigned char **adr, long *label)
{
  long sign = 1L;
  int status;
  (*adr)++;
  if (**adr == '-') {
    sign = -1L;
    (*adr)++;
  }
  if (isdigit(**adr)) {
    status = read_long(adr, label);
    if (status < 0) { return status; }
  }
  else { return E_R_DELTA_TICK; }

  label[0] = label[0] * sign;

  status = read_skip(adr);
  if (status == FIND_NULL) { return ADD_PREVIOUS_TIME; }

  return E_R_REMAIN_TIMEDATA;
}

static int read_mbt(unsigned char **adr, long *label)
{
  int count = 0;
  int status;

  while (**adr != '\0') {
    if (isdigit(**adr)) {
      status = read_long(adr, &label[count]);
      if (status < 0) { return status; }

      count++;
      if (count == 3) {
        status = read_skip(adr);
        if (status == FIND_NULL) { return count; }
        return E_R_REMAIN_TIMEDATA;
      }
    }
    else if (**adr == '-') {
      (*adr)++;

      if (!isdigit(**adr)) { return E_R_TIME_CHAR; }
      status = read_long(adr, &label[count]);
      if (status < 0) { return status; }

      label[count] = label[count] * (-1L);
      count++;

      status = read_skip(adr);
      if (status == FIND_NULL) { return count; }
      return E_R_REMAIN_TIMEDATA;
    }
    else { return E_R_TIME_CHAR; }

    status = read_skip(adr);
    if (status == FIND_NULL) { return count; }
  }

  return E_R_MBT_CHAR;
}

static int read_skip(unsigned char **adr)
{
  while (**adr != '\0') {
    if (!isskip(**adr)) {
      return FIND_CHAR;
    }
    (*adr)++;
  }

  (*adr)++;

  return FIND_NULL;
}

static int make_mbt(struct rmbt_st *inf, const int count, const long *label)
{
  int status;

  if (count == 3) {
    if (label[1] == 0) { return E_R_BEAT_VALUE; }

    status = make_bar(inf, label);
    if (status < 0) { return status; }

    inf->mbt[0] = label[0];
    inf->mbt[1] = label[1];
    inf->mbt[2] = label[2];
    return STATUS_SUCCESS;
  }
  if (count == 2) {
    if (label[0] == 0) { return E_R_BEAT_VALUE; }

    inf->mbt[1] = label[0];
    inf->mbt[2] = label[1];
    return STATUS_SUCCESS;
  }
  if (count == 1) {
    inf->mbt[2] = label[0];
    return STATUS_SUCCESS;
  }
  if (count == 0) {
    return STATUS_SUCCESS;
  }
  if (count == ADD_PREVIOUS_TIME) {
    inf->mbt[2] += label[0];
    return STATUS_SUCCESS;
  }

  return E_R_MAKE_MBT;
}

static int make_bar(struct rmbt_st *inf, const long *label)
{
  struct rbar_st *rrbar;

  while (label[0] > inf->max) {
    if (inf->max + 1 >= inf->lim) {
      inf->lim += MEAS_COUNT_STEP;
      if (inf->lim <= 0) { return E_R_BAR_COUNT; }
      rrbar = (struct rbar_st *) realloc(inf->bar,
                                         inf->lim * sizeof(struct rbar_st));
      if (rrbar == NULL) { return E_R_BAR_REALLOC; }
      inf->bar= rrbar;
    }

    inf->bar[inf->max + 1].tick =
      inf->bar[inf->max].tick +
      inf->bar[inf->max].beat.num * inf->bar[inf->max].beat.den;
    inf->bar[inf->max + 1].beat.num = inf->bar[inf->max].beat.num;
    inf->bar[inf->max + 1].beat.den = inf->bar[inf->max].beat.den;
    (inf->max)++;

    if (inf->bar[inf->max].tick <= 0) { return E_R_MEAS_TICK; }
  }

  return STATUS_SUCCESS;

}

static int store_mbt(struct rmbt_st *inf)
{
  inf->block[inf->blk].time =
    inf->bar[inf->mbt[0]].tick +
    inf->bar[inf->mbt[0]].beat.den * (inf->mbt[1] - 1) +
    inf->mbt[2];

  if (inf->block[inf->blk].time < 0) { return E_R_MINUS_TICK; }

  return STATUS_SUCCESS;
}

static int read_event(struct revt_st *buf)
{
  int type;
  int cnt;

  cnt = read_char(buf);
  if (cnt < 0) { return cnt; }
  if (cnt == FIND_NULL) { return E_R_EVENT_DATA; }
  type = (int) (*(buf->top) & 0xF0);

  if (type == 0x90) { return read_note(buf); }
  if (type == 0xB0) { return read_ddat(buf); }
  if (type == 0xF0) { return read_fxmg(buf); }
  if (type == 0xE0) { return read_ddat(buf); }
  if (type == 0xA0) { return read_ddat(buf); }
  if (type == 0x80) { return read_ddat(buf); }
  return read_sdat(buf);
}

static int read_note(struct revt_st *buf)
{
  int status;

  while (buf->cnt < 3) {
    status = read_char(buf);
    if (status < 0) { return status; }
    if (status == FIND_NULL) { break; }
  }
  if (buf->cnt != 3) { return E_R_DATA_LEN; }

  status = search_char(&(buf->in));
  if (status < 0) { return status; }
  if (status == FIND_NULL) {
    buf->dur = NONE_DUR;
    return STATUS_SUCCESS;
  }

  if (!isdigit(*(buf->in))) { return E_R_DUR_VALUE; }
  status = read_long(&(buf->in), &(buf->dur));
  if (status < 0) { return status; }

  status = read_char(buf);
  if (status < 0) { return status; }
  if (status == FIND_NULL) {
      *(buf->out) = 0;
      (buf->out)++;
      *(buf->out) = *(buf->top);
      (buf->out)++;
      return STATUS_SUCCESS;
  }

  status = search_char(&(buf->in));
  if (status != FIND_NULL) { return E_R_REMAIN_EVENT; }

  if ((buf->cnt) != 4) { return E_R_DATA_LEN; }

  (buf->cnt)--;

  *(buf->out) = (unsigned char) (*(buf->top) & 0x8F);
  (buf->out)++;

  return STATUS_SUCCESS;
}

static int read_sdat(struct revt_st *buf)
{
  int status;

  while ((buf->cnt) < 2) {
    status = read_char(buf);
    if (status < 0) { return status; }
    if (status == FIND_NULL) { break; }
  }
  if ((buf->cnt) != 2) { return E_R_DATA_LEN; }

  status = search_char(&(buf->in));
  if (status != FIND_NULL) { return E_R_REMAIN_EVENT; }

  return STATUS_SUCCESS;
}

static int read_ddat(struct revt_st *buf)
{
  int status;

  while ((buf->cnt) < 3) {
    status = read_char(buf);
    if (status < 0) { return status; }
    if (status == FIND_NULL) { break; }
  }
  if ((buf->cnt) != 3) { return E_R_DATA_LEN; }

  status = search_char(&(buf->in));
  if (status != FIND_NULL) { return E_R_REMAIN_EVENT; }

  return STATUS_SUCCESS;
}

static int read_fxmg(struct revt_st *buf)
{
  if (*(buf->top) == 0xFF) { return read_meta(buf); }
  if (*(buf->top) == 0xF0) { return read_syse(buf); }
  if (*(buf->top) == 0xF7) { return read_syse(buf); }

  return E_R_FXMG_VALUE;
}

static int read_syse(struct revt_st *buf)
{
  int len;
  int cnt;
  int status;

  do {
    status = read_char(buf);
    if (status < 0) { return status; }
  } while (status != FIND_NULL);

  cnt = read_chain(buf->top + 1);
  if (cnt < 0) { return cnt; }
  if (1 + cnt > buf->cnt) { return E_R_DATA_LEN; }

  len = read_len(buf->top + 1, cnt);
  if (buf->cnt != 1 + cnt + len) { return E_R_DATA_LEN; }

  return STATUS_SUCCESS;
}

static int read_meta(struct revt_st *buf)
{
  int len;
  int cnt;
  int status;

  do {
    status = read_char(buf);
    if (status < 0) { return status; }
  } while (status != FIND_NULL);

  cnt = read_chain(buf->top + 2);
  if (cnt < 0) { return cnt; }
  if (2 + cnt > buf->cnt) { return E_R_DATA_LEN; }

  len = read_len(buf->top + 2, cnt);
  if ((buf->cnt) != 2 + cnt + len) { return E_R_DATA_LEN; }

  return STATUS_SUCCESS;
}

static int read_chain(const unsigned char *data)
{
  if ((data[0] & 0x80) == 0) { return 1; }
  if ((data[1] & 0x80) == 0) { return 2; }
  if ((data[2] & 0x80) == 0) { return 3; }
  if ((data[3] & 0x80) == 0) { return 4; }

  return E_R_CHAIN_DATA;
}

static int read_len(const unsigned char *len, const int cnt)
{
  int value;

  if (cnt == 1) { return (int) len[0]; }

  value = (int) (len[0] & 0x7F) << 7;
  if (cnt == 2) { return value | (int) len[1]; }

  value = (value | ((int) (len[1] & 0x7F))) << 7;
  if (cnt == 3) { return value | (int) len[2]; }

  value = (value | ((int) (len[2] & 0x7F))) << 7;
  if (cnt == 4) { return value | (int) len[3]; }

  return E_R_LEN_DATA;
}

static int read_long(unsigned char **adr, long *number)
{
  int status;

  *number = (**adr) - '0';
  (*adr)++;
  while (isdigit(**adr)) {
    status = read_carry(number, **adr);
    if (status < 0) { return status; }
    (*adr)++;
  }

  return STATUS_SUCCESS;
}

static int read_carry(long *number, const unsigned char digit)
{
  long up;
  long lw;
  long over;
  const long hsize = sizeof(long) << 2;
  const long hmask = ~0UL >> (sizeof(long) << 2);

  up = *number >> hsize;
  lw = *number & hmask;
  lw = lw * 10 + (digit - '0');
  up = up * 10;
  over = (up + (lw >> hsize)) >> (hsize - 1);
  if (over) { return E_R_DATA_VALUE; }

  *number = (up << hsize) + lw;

  return STATUS_SUCCESS;
}

static int read_char(struct revt_st *buf)
{
  int status;

  status = search_char(&(buf->in));
  if (status < 0) { return status; }
  if (status == FIND_NULL) { return FIND_NULL; }

  if (*(buf->in) == '$')   { return read_hex(buf); }
  if (isdigit(*(buf->in))) { return read_value(buf); }
  if (istone(*(buf->in)))  { return read_tone(buf); }
  if (*(buf->in) == '^')   { return read_pitch(buf); }
  if (*(buf->in) == '@')   { return read_tempo(buf); }
  if (*(buf->in) == '\"')  { return read_text(buf); }
  if (*(buf->in) == '-')   { return read_minus(buf); }
  if (*(buf->in) == '/')   { return read_den(buf); }

  return E_R_PREFIX_TYPE;
}

static int istone(const unsigned char tone)
{
  if ('A' <= tone && tone <= 'H') { return 1; }
  if ('a' <= tone && tone <= 'h') { return 1; }

  return 0;
}

static int read_hex(struct revt_st *buf)
{
  int len = 0;
  int status;

  (buf->in)++;
  while (*(buf->in) != '\0') {
    status = read_upper(buf);
    if (status < 0) { return status; }
    if (status == HEX_END) { break; }

    status = read_lower(buf);
    if (status < 0) { return status; }

    len++;
    if (len < 0) { return E_R_HEX_VALUE; }
  }
  if (len == 0) { return E_R_HEX_DATA; }

  buf->cnt += len;
  if (buf->cnt <= 0) { return E_R_EVENT_BUF; }

  return len;
}

static int read_upper(struct revt_st *buf)
{
  if (isdigit(*(buf->in))) {
    *(buf->out) = (unsigned char) ((*(buf->in) - '0') << 4);
  }
  else if ('A' <= *(buf->in) && *(buf->in) <= 'F') {
    *(buf->out) = (unsigned char) ((*(buf->in) - 'A' + 10) << 4);
  }
  else if ('a' <= *(buf->in) && *(buf->in) <= 'f') {
    *(buf->out) = (unsigned char) ((*(buf->in) - 'a' + 10) << 4);
  }
  else {
    return HEX_END;
  }

  (buf->in)++;

  return STATUS_SUCCESS;
}

static int read_lower(struct revt_st *buf)
{
  if (isdigit(*(buf->in))) {
    *(buf->out) |= (unsigned char) (*(buf->in) - '0');
  }
  else if ('A' <= *(buf->in) && *(buf->in) <= 'F') {
    *(buf->out) |= (unsigned char) (*(buf->in) - 'A' + 10);
  }
  else if ('a' <= *(buf->in) && *(buf->in) <= 'f') {
    *(buf->out) |= (unsigned char) (*(buf->in) - 'a' + 10);
  }
  else if (*(buf->in) == 'x' || *(buf->in) == 'X') {
    if ((buf->cnt) != 0) { return E_R_SET_CHANNEL; }
    if ((buf->ch) & 0xF0) { return E_R_FIX_CHANNEL; }
    if (*(buf->out) == 0xF0) { return E_R_CHANNEL_VALUE; }
    *(buf->out) |= (unsigned char) (buf->ch);
  }
  else {
    return E_R_HEX_CHARACTER;
  }

  (buf->in)++;
  (buf->out)++;

  return STATUS_SUCCESS;
}

static int read_value(struct revt_st *buf)
{
  int value;

  value = read_num(&(buf->in), 255);
  if (value < 0) { return value; }

  *(buf->out) = (unsigned char) value;
  (buf->out)++;
  buf->cnt += 1;
  return 1;
}

static int read_tone(struct revt_st *buf)
{
  unsigned char tone;
  int status;

  status = read_step(*(buf->in));
  if (status < 0) { return status; }

  tone = (unsigned char) status;
  (buf->in)++;

  while (*(buf->in) != '\0') {
    if (*(buf->in) == 'b') { tone--; }
    else if (*(buf->in) == '#') { tone++; }
    else if (*(buf->in) == 'x') { tone = (unsigned char) (tone + 2); }
    else { break; }

    (buf->in)++;
  }

  if (*(buf->in) == '-') {
    (buf->in)++;
    if (isdigit(*(buf->in))) {
      if (*(buf->in) != '1') { return E_R_TONE_LOWER; }

      (buf->in)++;
    }
    tone = (unsigned char) (tone - 12);
  }
  else if (isdigit(*(buf->in))) {
    tone = (unsigned char) (tone + 12 * (*(buf->in) - '0'));
    (buf->in)++;
  }
  else {
    return E_R_TONE_NUMBER;
  }

  if (tone & 0x80) { return E_R_TONE_UPPER; }

  *(buf->out) = tone;
  (buf->out)++;
  buf->cnt += 1;
  return 1;
}

static int read_step(const unsigned char in)
{
  const int step[] = { 21, 23, 12, 14, 16, 17, 19, 23 };
  int tone;

  if ('A' <= in && in <= 'G') {
    tone = (int) in - 'A';
  }
  else if ('a' <= in && in <= 'g') {
    tone = (int) in - 'a';
  }
  else if (in == 'H') {
    tone = (int) ('B' - 'A');
  }
  else if (in == 'h') {
    tone = (int) ('b' - 'a');
  }
  else {
    return E_R_TONE_NAME;
  }

  return step[tone];
}

static int read_pitch(struct revt_st *buf)
{
  int value;

  (buf->in)++;

  if (*(buf->in) == '-') {
    (buf->in)++;
    if (!isdigit(*(buf->in))) { return E_R_PITCH_VALUE; }
    value = read_num(&(buf->in), 8192);
    if (value < 0) { return value; }
    value *= -1;
  }
  else {
    if (!isdigit(*(buf->in))) { return E_R_PITCH_VALUE; }
    value = read_num(&(buf->in), 8191);
    if (value < 0) { return value; }
  }

  value += 0x2000;
  *(buf->out) = (unsigned char) (value & 0x7F);
  (buf->out)++;
  *(buf->out) = (unsigned char) ((value >> 7) & 0x7F);
  (buf->out)++;
  (buf->cnt) += 2;
  return 2;
}

static int read_tempo(struct revt_st *buf)
{
  int value;

  (buf->in)++;

  if (!isdigit(*(buf->in))) { return E_R_TEMPO_VALUE; }

  value = read_num(&(buf->in), 60000000);
  if (value <  0) { return value; }
  if (value == 0) { return E_R_TEMPO_FAULT; }
  if (value <= 3) { return E_R_TEMPO_MIN; }

  value = 60000000 / value;

  *(buf->out) = (unsigned char) ((value >> 16) & 0xFF);
  (buf->out)++;
  *(buf->out) = (unsigned char) ((value >> 8) & 0xFF);
  (buf->out)++;
  *(buf->out) = (unsigned char) (value & 0xFF);
  (buf->out)++;
  (buf->cnt) += 3;
  return 3;
}

static int read_text(struct revt_st *buf)
{
  int cnt = 0;
  int len;
  int status;
  unsigned char *str;

  str = (buf->out) + 1;
  (buf->in)++;
  while (*(buf->in) != '\"') {
    if (*(buf->in) == '\0') { return E_R_STR_CLOSE; }

    if (*(buf->in) == '\\') {
      (buf->in)++;
      status = read_esc(buf->in);
      if (status < 0) { return status; }
    }
    *str = *(buf->in);
    cnt++;
    (buf->in)++;
    str++;
  }

  (buf->in)++;

  status = shift_text(cnt, buf->out + 1);
  if (status < 0) { return status; }

  len = read_var(cnt, &(buf->out));
  if (len < 0) { return len; }

  buf->out = buf->out + cnt;

  status = search_char(&(buf->in));
  if (status != FIND_NULL) { return E_R_REMAIN_EVENT; }

  (buf->cnt) += len + cnt;
  return len + cnt;
}

static int read_esc(unsigned char *in)
{
  if (*in == '\0') { return E_R_STR_CLOSE; }

  if (*in == 'n') { *in = '\n'; return STATUS_SUCCESS; }
  if (*in == 'r') { *in = '\r'; return STATUS_SUCCESS; }
  if (*in == 't') { *in = '\t'; return STATUS_SUCCESS; }
  if (*in == 'v') { *in = '\v'; return STATUS_SUCCESS; }
  if (*in == 'f') { *in = '\f'; return STATUS_SUCCESS; }
  if (*in == 'b') { *in = '\b'; return STATUS_SUCCESS; }
  if (*in == 'a') { *in = '\a'; return STATUS_SUCCESS; }
  return STATUS_SUCCESS;
}

static int shift_text(const int cnt, unsigned char *str)
{
  int shift;

  if (cnt < 0x80) { return STATUS_SUCCESS; }
  else if (cnt < 0x4000) { shift = 1; }
  else if (cnt < 0x200000) { shift = 2; }
  else if (cnt < 0x10000000) { shift = 3; }
  else { return E_R_TEXT_LEN; }

  memmove(str + shift, str, (size_t) cnt);

  return STATUS_SUCCESS;
}

static int read_var(int cnt, unsigned char **adr)
{
  if (cnt < 0x80) {
    **adr = (unsigned char) cnt;
    (*adr)++;
    return 1;
  }
  if (cnt < 0x4000) {
    **adr = (unsigned char) ((cnt >> 7) | 0x80);
    (*adr)++;
    **adr = (unsigned char) (cnt & 0x7F);
    (*adr)++;
    return 2;
  }
  if (cnt < 0x200000) {
    **adr = (unsigned char) ((cnt >> 14) | 0x80);
    (*adr)++;
    **adr = (unsigned char) ((cnt >> 7) | 0x80);
    (*adr)++;
    **adr = (unsigned char) (cnt & 0x7F);
    (*adr)++;
    return 3;
  }
  if (cnt < 0x10000000) {
    **adr = (unsigned char) ((cnt >> 21) | 0x80);
    (*adr)++;
    **adr = (unsigned char) ((cnt >> 14) | 0x80);
    (*adr)++;
    **adr = (unsigned char) ((cnt >> 7) | 0x80);
    (*adr)++;
    **adr = (unsigned char) (cnt & 0x7F);
    (*adr)++;
    return 4;
  }
  return E_R_TEXT_LEN;
}

static int read_minus(struct revt_st *buf)
{
  int value;

  (buf->in)++;
  if (!isdigit(*(buf->in))) { return E_R_MINUS_VALUE; }
  value = read_num(&(buf->in), 128);
  if (value < 0) { return value; }

  *(buf->out) = (unsigned char) (((~value) + 1) & 0xFF);
  (buf->out)++;
  buf->cnt += 1;
  return 1;
}

static int read_den(struct revt_st *buf)
{
  int den;

  (buf->in)++;

  den = read_exp(buf);
  if (den < 0) { return den; }

  *(buf->out) = (unsigned char) den;
  (buf->out)++;
  buf->cnt += 1;

  return 1;
}

static int read_exp(struct revt_st *buf)
{
  int i;
  int crr;
  int rem;
  int cnt;
  int off = 0;
  int len;
  unsigned char *data;
  data = buf->in;

  len = read_pgrs(buf);
  if (len <= 0) { return E_R_BEAT_DATA; }

  for (cnt = 0; cnt < 0x80; cnt++) {
    rem = 0;
    for (i = off; i< len; i++) {
      crr = (int) data[i];
      data[i] = (unsigned char) ((10 * rem + data[i]) >> 1);
      rem = crr & 1;
    }
    if (rem != 0) { return cnt; }
    if (data[off] == 0) { off++; }
  }
  return E_R_BEAT_UPPER;
}

static int read_pgrs(struct revt_st *buf)
{
  int len = 0;
  int or = 0;

  while (*(buf->in) != '\0') {

    if (isdigit(*(buf->in))) {
      *(buf->in) = (unsigned char)(*(buf->in) - '0');
    }
    else {
      break;
    }

    or = or | *(buf->in);
    len++;
    (buf->in)++;
  }
  if (or == 0) { return 0; }
  return len;
}

static int read_num(unsigned char **adr, int max)
{
  int value;
  value = **adr - '0';
  (*adr)++;
  while (isdigit(**adr)) {
    value = value * 10 + ((**adr) - '0');
    (*adr)++;
    if (value > max) { return E_R_NUM_VALUE; }
  }
  return value;
}

static int search_char(unsigned char **adr)
{
  while (**adr != '\0') {
    if (!isskip(**adr)) {
      return FIND_CHAR;
    }
    (*adr)++;
  }

  return FIND_NULL;
}

static int isskip(const unsigned char c)
{
  if (c == ' ') { return 1; }
  if (c == '\t') { return 1; }
  if (c == ',') { return 1; }
  return 0;
}

static int store_ch(struct revt_st *buf)
{
  if ((*(buf->top) & 0xF0) == 0xF0) { return STATUS_SUCCESS; }

  buf->ch = (int) (*(buf->top) & 0x0F);

  return STATUS_SUCCESS;
}

static int store_den(struct midi_st *midi, struct rmbt_st *inf)
{
  if (inf->trk != 0) { return STATUS_SUCCESS; }

  if (inf->buf.top[0] != 0xFF) { return STATUS_SUCCESS; }
  if (inf->buf.top[1] != 0x58) { return STATUS_SUCCESS; }

  if (inf->mbt[1] != 1) { return STATUS_SUCCESS; }
  if (inf->mbt[2] != 0) { return STATUS_SUCCESS; }

  if (inf->mbt[0] < inf->max) { return E_R_METER_CHANGE; }

  inf->bar[inf->mbt[0]].beat.num = (int) inf->buf.top[3];
  if (inf->bar[inf->mbt[0]].beat.num <= 0) { return E_R_SET_BEAT; }

  if (inf->buf.top[4] >= 2) {
    inf->bar[inf->mbt[0]].beat.den =
      midi->header.tpqn >> ((int) inf->buf.top[4] - 2);
  }
  else {
    inf->bar[inf->mbt[0]].beat.den =
      midi->header.tpqn << (2 - (int) inf->buf.top[4]);
  }

  if (inf->bar[inf->mbt[0]].beat.den <= 0) { return E_R_SET_BEAT; }

  return STATUS_SUCCESS;
}

static int store_event(struct rmbt_st *inf)
{
  inf->block[inf->blk].event =
    (unsigned char *) malloc(inf->buf.cnt * sizeof(unsigned char));

  if (inf->block[inf->blk].event == NULL) { return E_R_EVENT_MALLOC; }

  memcpy(inf->block[inf->blk].event, inf->buf.top, (size_t) inf->buf.cnt);

  return STATUS_SUCCESS;
}

static int store_mute(struct rmbt_st *inf)
{
  int status;
  long ntime;

  if ((inf->buf.top[0] & 0xF0) != 0x90) { return STATUS_SUCCESS; }
  if (inf->buf.dur == NONE_DUR) { return STATUS_SUCCESS; }

  ntime = inf->block[inf->blk].time + inf->buf.dur;
  if (ntime < 0) { return E_R_MUTE_DUR; }

  status = store_block(inf);
  if (status < 0) { return status; }

  status = realloc_block(inf);
  if (status < 0) { return status; }

  inf->block[inf->blk].time = ntime;

  inf->block[inf->blk].event =
    (unsigned char *) malloc(3 * sizeof(unsigned char));

  if (inf->block[inf->blk].event == NULL) { return E_R_EVENT_MALLOC; }

  inf->block[inf->blk].event[0] = inf->buf.top[4];
  inf->block[inf->blk].event[1] = inf->buf.top[1];
  inf->block[inf->blk].event[2] = inf->buf.top[3];

  return STATUS_SUCCESS;
}

static int store_block(struct rmbt_st *inf)
{
  int status;

  status = sort_block(inf->block, inf->blk);
  if (status < 0) { return status; }

  (inf->blk)++;
  if (inf->blk < 0) { return E_R_BLOCK_COUNT; }

  return STATUS_SUCCESS;
}

static int sort_block(struct block_st *block, int blk)
{
  struct block_st swp;

  while (blk > 0) {
    if (block[blk].time >= block[blk - 1].time) {
      return STATUS_SUCCESS;
    }
    swp.event = block[blk].event;
    block[blk].event = block[blk - 1].event;
    block[blk - 1].event = swp.event;

    swp.time = block[blk].time;
    block[blk].time = block[blk - 1].time;
    block[blk - 1].time = swp.time;

    blk--;
  }

  return STATUS_SUCCESS;
}

static int store_eot(struct midi_st *midi, struct rmbt_st *inf)
{
  int i;
  int status;
  struct track_st *rtrack;

  if (inf->buf.top[1] != 0x2F) { return STATUS_SUCCESS; }
  if (inf->buf.top[0] != 0xFF) { return STATUS_SUCCESS; }

  if (inf->trk >= 0xFFFF) { return E_R_TRACK_COUNT; }

  if (inf->trk >= inf->eot) {
    inf->eot += TRACK_COUNT_STEP;
    rtrack =
      (struct track_st *) realloc(midi->track,
                                  inf->eot * sizeof(struct track_st));
    if (rtrack == NULL) { return E_R_TRACK_REALLOC; }
    midi->track = rtrack;

    status = init_rtrk(midi->track, inf->trk, inf->eot);
    if (status < 0) { return status; }
  }

  midi->track[inf->trk].block =
    (struct block_st *) malloc(((unsigned int) inf->blk + 1U)
                               * sizeof(struct block_st));
  if (midi->track[inf->trk].block == NULL) { return E_R_BLOCK_MALLOC; }

  for (i = 0; i < inf->blk; i++) {
    midi->track[inf->trk].block[i].time = inf->block[i].time;
    inf->block[i].time = -1;
    midi->track[inf->trk].block[i].event = inf->block[i].event;
    inf->block[i].event = NULL;
  }
  midi->track[inf->trk].block[i].time = -1;
  midi->track[inf->trk].block[i].event = NULL;

  (inf->trk)++;
  inf->blk = 0;

  status = init_mbt(inf->mbt);
  if (status < 0) { return status; }

  inf->buf.ch = 0xFF;

  return STATUS_SUCCESS;
}

static int free_inf(struct rmbt_st *inf)
{
  int i;

  if (inf->block != NULL) {
    for (i = 0; i < inf->eob; i++) {
      free(inf->block[i].event);
    }
    free(inf->block);
  }

  if (inf->buf.top != NULL) { free(inf->buf.top); }
  if (inf->bar != NULL) { free(inf->bar); }

  return STATUS_SUCCESS;
}
