/* Copyright (c) 2013 tonomu
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include <stdio.h>
#include <stdlib.h>
#include "mbt.h"

static int write_header(struct header_st *header, FILE *ofp);
static int init_winf(const int tpqn, struct wmbt_st *inf);
static int write_track(struct block_st *block, struct wmbt_st *inf, FILE *ofp);
static int init_track(struct wmbt_st *inf, long *label, FILE *ofp);
static int write_tcom(const int trk, FILE *ofp);
static int ismute(const unsigned char *event);
static int write_mbt(long time, struct wmbt_st *inf, long *label, FILE *ofp);
static int store_label(const long *mbt, long *label);
static int store_meas(const long time, struct wmbt_st *inf);
static int store_bt(const long time, struct wmbt_st *inf);
static int write_label(const long *mbt, const long *label, FILE *ofp);
static int write_event(struct block_st *block, struct wmbt_st *inf, FILE *ofp);
static int write_ins(struct block_st *block, struct wmbt_st *inf, FILE *ofp);
static int write_note(struct block_st *block, const int key, FILE *ofp);
static int write_ctrc(const unsigned char *event, FILE *ofp);
static int write_fxmg(const unsigned char *event, FILE *ofp);
static int write_bend(const unsigned char *event, FILE *ofp);
static int write_pres(const unsigned char *event, FILE *ofp);
static int write_prgc(const unsigned char *event, FILE *ofp);
static int write_poly(const unsigned char *event, const int key, FILE *ofp);
static int write_tone(const unsigned char tone, const int key, FILE *ofp);
static int search_mute(struct block_st *block, int *mute);
static int write_drum(const unsigned char data, FILE *ofp);
static int write_ccom(const unsigned char ctrl, FILE *ofp);
static int write_meta(const unsigned char *event, FILE *ofp);
static int write_syse(const unsigned char *event, FILE *ofp);
static int check_len(const unsigned char *event, int *value);
static int write_mcom(const unsigned char *data, int type, int len, FILE *ofp);
static int write_mpad(const int len, FILE *ofp);
static int write_text(const unsigned char *data, int len, FILE *ofp);
static int write_tempo(const unsigned char *data, FILE *ofp);
static int write_tsig(const unsigned char *data, FILE *ofp);
static int write_exp(int cnt, FILE *ofp);
static int write_ksig(const unsigned char *data, FILE *ofp);
static int write_spad(const int len, FILE *ofp);
static int write_dcom(const unsigned char prog, FILE *ofp);
static int write_pcom(const unsigned char prog, FILE *ofp);
static int make_bar(const unsigned char *event, struct wmbt_st *inf);

enum {
  MEAS_COUNT_STEP = 512,
  NOTE_OFFSET = 12,
  NOTE_OFF = 1,
  VEL_ZERO = 2,
  NONE_MUTE = 0,
};

static const char * const DRUM_COMMENT[] = {
"Acoustic Bass Drum",
"Bass Drum 1",
"Side Stick",
"Snare 1",
"Hand Clap",
"Snare 2",
"Low Floor Tom",
"Closed Hi-hat",
"High Floor Tom",
"Pedal Hi-hat",
"Low Tom",
"Open Hi-hat",
"Low-Mid Tom",
"High-Mid Tom",
"Crash Cymbal 1",
"High Tom",
"Ride Cymbal 1",
"Chinese Cymbal",
"Ride Bell",
"Tambourine",
"Splash Cymbal",
"Cowbell",
"Crash Cymbal 2",
"Vibra-slap",
"Ride Cymbal 2",
"High Bongo",
"Low Bongo",
"Mute Hi Conga",
"Open Hi Conga",
"Low Conga",
"High Timbale",
"Low Timbale",
"High Agogo",
"Low Agogo",
"Cabasa",
"Maracas",
"Short Whistle",
"Long Whistle",
"Short Guiro",
"Long Guiro",
"Claves",
"Hi Wood Block",
"Low Wood Block",
"Mute Cuica",
"Open Cuica",
"Mute Triangle",
"Open Triangle",
"Undefined Drum Sound"
};

static const char * const CTRL_COMMENT[] = {
"Bank Select MSB",
"Modulation Wheel MSB",
"Breath Controller MSB",
"Control Change 3",
"Foot Controller MSB",
"Portamento Time MSB",
"Data Entry MSB",
"Channel Volume MSB",
"Balance MSB",
"Control Change 9",
"Pan MSB",
"Expression Controller MSB",
"Effect Control 1 MSB",
"Effect Control 2 MSB",
"Control Change 14",
"Control Change 15",
"General Purpose Ctrl 1 MSB",
"General Purpose Ctrl 2 MSB",
"General Purpose Ctrl 3 MSB",
"General Purpose Ctrl 4 MSB",
"Control Change 20",
"Control Change 21",
"Control Change 22",
"Control Change 23",
"Control Change 24",
"Control Change 25",
"Control Change 26",
"Control Change 27",
"Control Change 28",
"Control Change 29",
"Control Change 30",
"Control Change 31",
"Bank Select LSB",
"Modulation Wheel LSB",
"Breath Controller LSB",
"Control Change 35",
"Foot Controller LSB",
"Portamento Time LSB",
"Data Entry LSB",
"Channel Volume LSB",
"Balance LSB",
"Control Change 41",
"Pan LSB",
"Expression Controller LSB",
"Effect Control 1 LSB",
"Effect Control 2 LSB",
"Control Change 46",
"Control Change 47",
"General Purpose Ctrl 1 LSB",
"General Purpose Ctrl 2 LSB",
"General Purpose Ctrl 3 LSB",
"General Purpose Ctrl 4 LSB",
"Control Change 52",
"Control Change 53",
"Control Change 54",
"Control Change 55",
"Control Change 56",
"Control Change 57",
"Control Change 58",
"Control Change 59",
"Control Change 60",
"Control Change 61",
"Control Change 62",
"Control Change 63",
"Damper Pedal on/off",
"Portamento On/Off",
"Sostenuto On/Off",
"Soft Pedal On/Off",
"Legato Footswitch",
"Freeze (Hold 2)",
"Sound Variation",
"Harmonic Intensity",
"Release Time",
"Attack Time",
"Brightness",
"Decay Time",
"Vibrato Rate",
"Vibrato Depth",
"Vibrato Delay",
"Sound Controller 10",
"General Purpose Controller 5",
"General Purpose Controller 6",
"General Purpose Controller 7",
"General Purpose Controller 8",
"Portamento Control",
"Control Change 85",
"Control Change 86",
"Control Change 87",
"Control Change 88",
"Control Change 89",
"Control Change 90",
"Reverb Send Level",
"Tremolo Depth",
"Chorus Send Level",
"Celeste",
"Phaser Depth",
"Data Increment",
"Data Decrement",
"NRPN LSB",
"NRPN MSB",
"RPN LSB",
"RPN MSB",
"Control Change 102",
"Control Change 103",
"Control Change 104",
"Control Change 105",
"Control Change 106",
"Control Change 107",
"Control Change 108",
"Control Change 109",
"Control Change 110",
"Control Change 111",
"Control Change 112",
"Control Change 113",
"Control Change 114",
"Control Change 115",
"Control Change 116",
"Control Change 117",
"Control Change 118",
"Control Change 119",
"Sound Off",
"Reset All Controllers",
"Local Control On/Off",
"All Notes Off",
"Omni Mode Off",
"Omni Mode On",
"Poly Mode On/Off",
"Poly Mode On"
};

static const char * const PROG_COMMENT[] = {
"Piano1",
"Piano2",
"Piano3",
"Honky-tonk",
"E.Piano1",
"E.Piano2",
"Harpsichord",
"Clavi",
"Celesta",
"Glockenspiel",
"Music Box",
"Vibraphone",
"Marimba",
"Xylophone",
"Tubular-bell",
"Santur",
"Organ1",
"Organ2",
"Organ3",
"Church Organ",
"Reed Organ",
"Accordion",
"Harmonica",
"Bandoneon",
"Acoustic Guitar (Nylon)",
"Acoustic Guitar (Steel)",
"E.Guitar (Jazz)",
"E.Guitar (Clean)",
"E.Guitar (Muted)",
"Overdrive Guitar",
"Distortion Guitar",
"Guitar Harmonics",
"Acoustic Bass",
"E.Bass (Finger)",
"E.Bass (Pick)",
"Fretless Bass",
"Slap Bass1",
"Slap Bass2",
"Synth Bass1",
"Synth Bass2",
"Violin",
"Viola",
"Cello",
"Contrabass",
"Tremolo Strings",
"Pizzicato Strings",
"Harp",
"Timpani",
"Strings Ensemble1",
"Strings Ensemble2",
"Synth Strings1",
"Synth Strings2",
"Choir Aahs",
"Voice Oohs",
"Synth Vox",
"Orchestra Hit",
"Trumpet",
"Trombone",
"Tuba",
"Muted Trumpet",
"French Horn",
"Brass Section",
"Synth Brass1",
"Synth Brass2",
"Soprano Sax",
"Alto Sax",
"Tenor Sax",
"Baritone Sax",
"Oboe",
"English Horn",
"Bassoon",
"Clarinet",
"Piccolo",
"Flute",
"Recorder",
"Pan Flute",
"Bottle Blow",
"Shakuhachi",
"Whistle",
"Ocarina",
"Square Wave",
"Saw Wave",
"Synth Calliope",
"Chiffer Lead",
"Charang",
"Solo Vox",
"5th Saw Wave",
"Bass & Lead",
"Fantasia",
"Warm Pad",
"Polysynth",
"Space Voice",
"Bowed Glass",
"Metal Pad",
"Halo Pad",
"Sweep Pad",
"Ice rain",
"Soundtrack",
"Crystal",
"Atmosphere",
"Brightness",
"Goblin",
"Echo Drops",
"Star Theme",
"Sitar",
"Banjo",
"Shamisen",
"Koto",
"Kalimba",
"Bagpipe",
"Fiddle",
"Shehnai",
"Tinkle Bell",
"Agogo",
"Steel Drum",
"Woodblock",
"Taiko",
"Melodic Tom",
"Synth Drum",
"Reverse Cymbal",
"Guitar Fret Noise",
"Breath Noise",
"Seashore",
"Bird",
"Telephone",
"Helicopter",
"Applause",
"Gun Shot"
};

int Write_Data(struct midi_st *midi ,FILE *ofp)
{
  struct wmbt_st inf;
  int status;

  if (midi->header.tpqn >= 0x8000) { return E_W_TPQN_VALUE; }
  if (midi->header.tpqn <= 0) { return E_W_TPQN_DATA; }

  status = write_header(&midi->header, ofp);
  if (status < 0) { return status; }

  status = init_winf(midi->header.tpqn, &inf);
  if (status < 0) { return status; }

  for (inf.trk = 0; inf.trk < midi->header.total; (inf.trk)++) {
    status = write_track(midi->track[inf.trk].block, &inf, ofp);
    if (status < 0) {
      free(inf.bar);
      return status;
    }
  }
  free(inf.bar);

  return STATUS_SUCCESS;
}

static int write_header(struct header_st *header, FILE *ofp)
{
  int status;

  status = fprintf(ofp, ";; Format : %d\n", header->format);
  if (status < 0) { return E_W_WRITE_HEADER; }

  status = fprintf(ofp, ";; Tracks : %d\n", header->total);
  if (status < 0) { return E_W_WRITE_HEADER; }

  status = fprintf(ofp, ";; TPQN   : %d\n", header->tpqn);
  if (status < 0) { return E_W_WRITE_HEADER; }

  return STATUS_SUCCESS;
}

static int init_winf(const int tpqn, struct wmbt_st *inf)
{
  inf->tpqn = tpqn;
  inf->trk = 0;
  inf->max = 0;
  inf->lim = MEAS_COUNT_STEP - 1;
  inf->bar = (struct wbar_st *) malloc(inf->lim * sizeof(struct wbar_st));
  if (inf->bar == NULL) { return E_W_BAR_MALLOC; }
  inf->bar[0].tick = 0;
  inf->bar[0].key = 0;
  inf->bar[0].beat.num = 4;
  inf->bar[0].beat.den = inf->tpqn;
  inf->bar[1].tick = 4 * inf->tpqn;

  return STATUS_SUCCESS;
}

static int write_track(struct block_st *block, struct wmbt_st *inf, FILE *ofp)
{
  int status;
  long label[3];

  status = init_track(inf, label, ofp);
  if (status < 0) { return status; }

  while (block->event != NULL) {
    if (!ismute(block->event)) {
      status = write_mbt(block->time, inf, label, ofp);
      if (status < 0) { return status; }
      status = write_event(block, inf, ofp);
      if (status < 0) { return status; }
    }
    status = make_bar(block->event, inf);
    if (status < 0) { return status; }

    block++;
  }
  return STATUS_SUCCESS;
}

static int init_track(struct wmbt_st *inf, long *label, FILE *ofp)
{
  int status;

  inf->mbt[0] = 0;
  inf->mbt[1] = 1;
  inf->mbt[2] = 0;
  label[0] = -1;
  inf->ch = 0xFF;

  status = write_tcom(inf->trk, ofp);
  if (status < 0) { return status; }

  return STATUS_SUCCESS;
}

static int write_tcom(const int trk, FILE *ofp)
{
  int status;
  status = fprintf(ofp,
                   "\n;=== Track %03d =================================\n",
                   trk);
  if (status < 0) { return E_W_TRACK_COM; }

  return STATUS_SUCCESS;
}

static int ismute(const unsigned char *event)
{
  if ((event[0] & 0xF0) == 0x80) { return 1; }
  if ((event[0] & 0xF0) != 0x90) { return 0; }
  if (event[2] == 0x00) { return 1; }

  return 0;
}

static int write_mbt(long time, struct wmbt_st *inf, long *label, FILE *ofp)
{
  int status;

  status = store_meas(time, inf);
  if (status < 0) { return status; }

  status = store_bt(time, inf);
  if (status < 0) { return status; }

  status = write_label(inf->mbt, label, ofp);
  if (status < 0) { return status; }

  status = store_label(inf->mbt, label);
  if (status < 0) { return status; }

  return STATUS_SUCCESS;
}

static int store_meas(const long time, struct wmbt_st *inf)
{
  struct wbar_st *rwbar;

  while (time >= inf->bar[inf->mbt[0] + 1].tick) {
    inf->mbt[0]++;

    if (inf->mbt[0] > inf->max) {
      do {
        if (inf->max + 2 >= inf->lim) {
          inf->lim += MEAS_COUNT_STEP;
          if (inf->lim <= 0) { return E_W_BAR_VALUE; }
          rwbar = (struct wbar_st *) realloc(inf->bar,
                                             inf->lim
                                             * sizeof(struct wbar_st));
          if (rwbar == NULL) { return E_W_BAR_REALLOC; }
          inf->bar = rwbar;
        }

        inf->bar[inf->max + 1].beat.num = inf->bar[inf->max].beat.num;
        inf->bar[inf->max + 1].beat.den = inf->bar[inf->max].beat.den;
        inf->bar[inf->max + 1].key = inf->bar[inf->max].key;
        inf->max++;
        inf->bar[inf->max + 1].tick =
          inf->bar[inf->max].tick +
          inf->bar[inf->max].beat.num * inf->bar[inf->max].beat.den;
        if (inf->bar[inf->max + 1].tick <= 0) { return E_W_MEAS_TICK; }
      } while (time >= inf->bar[inf->max + 1].tick);
      inf->mbt[0] = inf->max;
      return STATUS_SUCCESS;
    }
  }

  return STATUS_SUCCESS;
}

static int store_bt(const long time, struct wmbt_st *inf)
{
  long off;
  long den;

  off = time - inf->bar[inf->mbt[0]].tick;
  den = inf->bar[inf->mbt[0]].beat.den;

  inf->mbt[1] = (off / den) + 1;
  inf->mbt[2] = (off % den);

  return STATUS_SUCCESS;
}

static int write_label(const long *mbt, const long *label, FILE *ofp)
{
  int status;

  if (mbt[0] != label[0]) {
    status = fprintf(ofp, "%4ld %2ld %5ld :", mbt[0], mbt[1], mbt[2]);
    if (status < 0) { return E_W_WRITE_MBT; }
    return STATUS_SUCCESS;
  }
  if (mbt[1] != label[1]) {
    status = fprintf(ofp, "     %2ld %5ld :", mbt[1], mbt[2]);
    if (status < 0) { return E_W_WRITE_MBT; }
    return STATUS_SUCCESS;
  }
  if (mbt[2] != label[2]) {
    status = fprintf(ofp, "        %5ld :", mbt[2]);
    if (status < 0) { return E_W_WRITE_MBT; }
    return STATUS_SUCCESS;
  }
  status = fprintf(ofp, "              :");
  if (status < 0) { return E_W_WRITE_MBT; }
  return STATUS_SUCCESS;
}

static int store_label(const long *mbt, long *label)
{
  label[0] = mbt[0];
  label[1] = mbt[1];
  label[2] = mbt[2];

  return STATUS_SUCCESS;
}

static int write_event(struct block_st *block, struct wmbt_st *inf, FILE *ofp)
{
  int status;
  int key = inf->bar[inf->mbt[0]].key;
  int ins = (int) (block->event[0] & 0xF0);

  status = write_ins(block, inf, ofp);
  if (status < 0) { return status; }

  if (ins == 0x90) { return write_note(block, key, ofp); }
  if (ins == 0xB0) { return write_ctrc(block->event, ofp); }
  if (ins == 0xF0) { return write_fxmg(block->event, ofp); }
  if (ins == 0xE0) { return write_bend(block->event, ofp); }
  if (ins == 0xD0) { return write_pres(block->event, ofp); }
  if (ins == 0xC0) { return write_prgc(block->event, ofp); }
  if (ins == 0xA0) { return write_poly(block->event, key, ofp); }

  return E_W_EVENT_TYPE;
}

static int write_ins(struct block_st *block, struct wmbt_st *inf, FILE *ofp)
{
  int ins = (int) block->event[0];
  int status;

  if ((ins & 0xF0) == 0xF0) {
    status = fprintf(ofp, " $%02X", (unsigned int) ins);
    if (status < 0) { return E_W_WRITE_INS; }
    return STATUS_SUCCESS;
  }

  if (inf->ch == (ins & 0x0F)) {
    status = fprintf(ofp, " $%01Xx", (unsigned int) (ins >> 4));
    if (status < 0) { return E_W_WRITE_INS; }
    return STATUS_SUCCESS;
  }

  inf->ch = ins & 0x0F;

  status = fprintf(ofp, " $%02X", (unsigned int) ins);
  if (status < 0) { return E_W_WRITE_INS; }

  return STATUS_SUCCESS;
}

static int write_note(struct block_st *block, const int key, FILE *ofp)
{
  int mute;
  int status;

  status = write_tone(block->event[1], key, ofp);
  if (status < 0) { return status; }

  status = fprintf(ofp, "  %3d", (int) block->event[2]);
  if (status < 0) { return E_W_WRITE_DATA; }

  status = search_mute(block, &mute);
  if (status < 0) { return E_W_WRITE_DATA; }

  if ((block->event[0] & 0x0F) == 0x09) {
    if (status == NOTE_OFF) {
      status = fprintf(ofp, " %6ld %4d",
                       block[mute].time - block[0].time,
                       (int) block[mute].event[2]);
    }
    else if (status == VEL_ZERO) {
      status = fprintf(ofp, " %6ld     ", block[mute].time - block[0].time);
    }
    else {
      status = fprintf(ofp, "            ");
    }
    if (status < 0) { return E_W_WRITE_DATA; }

    status = write_drum(block->event[1], ofp);
    if (status < 0) { return status; }
  }
  else {
    if (status == NOTE_OFF) {
      status = fprintf(ofp,
                       " %6ld %4d\n",
                       block[mute].time - block[0].time,
                       (int) block[mute].event[2]);
    }
    else if (status == VEL_ZERO) {
      status = fprintf(ofp, " %6ld\n", block[mute].time - block[0].time);
    }
    else {
      status = fprintf(ofp, "\n");
    }
    if (status < 0) { return E_W_WRITE_DATA; }
  }

  return STATUS_SUCCESS;
}

static int write_ctrc(const unsigned char *event, FILE *ofp)
{
  int status;
  status = fprintf(ofp, "  %3d  %3d", (int) event[1], (int) event[2]);
  if (status < 0) { return E_W_WRITE_DATA; }

  status = write_ccom(event[1], ofp);
  if (status < 0) { return E_W_WRITE_DATA; }

  return STATUS_SUCCESS;
}

static int write_fxmg(const unsigned char *event, FILE *ofp)
{
  if (*event == 0xFF) { return write_meta(event, ofp); }
  if (*event == 0xF0) { return write_syse(event, ofp); }
  if (*event == 0xF7) { return write_syse(event, ofp); }

  return E_W_EVENT_TYPE;
}

static int write_bend(const unsigned char *event, FILE *ofp)
{
  int status;
  int bend;

  if ((event[1] | event[2]) >> 7) { return E_W_BEND_DATA; }

  bend = ((event[2] << 7) | event[1]) - 0x2000;

  status = fprintf(ofp,
                   "  ^%-5d                     ;Pitch Bend Change\n",
                   bend);
  if (status < 0) { return E_W_WRITE_DATA; }

  return STATUS_SUCCESS;
}

static int write_pres(const unsigned char *event, FILE *ofp)
{
  int status;

  status = fprintf(ofp,
                   "  %3d                        ;Channel Pressure\n",
                   (int) event[1]);

  if (status < 0) { return E_W_WRITE_DATA; }

  return STATUS_SUCCESS;
}

static int write_prgc(const unsigned char *event, FILE *ofp)
{
  int status;

  status = fprintf(ofp, "  %3d                        ", (int) event[1]);
  if (status < 0) { return E_W_WRITE_DATA; }

  if ((event[0] & 0x0F) == 0x09) {
    status = write_dcom(event[1], ofp);
  }
  else {
    status = write_pcom(event[1], ofp);
  }

  if (status < 0) { return status; }

  return STATUS_SUCCESS;
}

static int write_poly(const unsigned char *event, const int key, FILE *ofp)
{
  int status;

  status = write_tone(event[1], key, ofp);
  if (status < 0) { return status; }

  status = fprintf(ofp,
                   "  %3d                   ;Polyphonic Key Pressure\n",
                   (int) event[2]);
  if (status < 0) { return E_W_WRITE_DATA; }

  return STATUS_SUCCESS;
}

static int write_tone(const unsigned char tone, const int key, FILE *ofp)
{
  int status;
  int base;
  int rest;
  const char * const name[] = {
    "C", "C#", "D", "D#", "E", "F", "F#", "G", "G#", "A", "A#", "B",
    "C", "Db", "D", "Eb", "E", "F", "Gb", "G", "Ab", "A", "Bb", "B"
  };
  const int len[24] = {
    0  ,  1  , 0  , 1   , 0  , 0  ,  1  ,  0 ,  1  ,  0 ,  1  , 0,
    0  ,  1  , 0  , 1   , 0  , 0  ,  1  ,  0 ,  1  ,  0 ,  1  , 0
  };
  const char * const oct[] = {
    "-", "0", "1", "2", "3", "4", "5", "6", "7", "8", "9"
  };

  base = (int) tone / 12;
  rest = (int) tone % 12;
  rest = rest + key;

  status = fprintf(ofp, "  %s%s", name[rest], oct[base]);
  if (status < 0) { return E_W_WRITE_TONE; }

  if (len[rest]) { return STATUS_SUCCESS; }

  status = fputs(" ", ofp);
  if (status == EOF) { return E_W_WRITE_TONE; }

  return STATUS_SUCCESS;
}

static int search_mute(struct block_st *block, int *mute)
{
  int channel;
  int note;

  channel = (int) block->event[0] & 0x0F;
  note = (int) block->event[1];
  block++;

  for ((*mute) = 1; block->event != NULL; (*mute)++) {
    if (block->event[1] == note) {
      if (block->event[0] == (0x80 | channel)) {
        return NOTE_OFF;
      }
      if (block->event[0] == (0x90 | channel)) {
        if (block->event[2] == 0x00) {
          return VEL_ZERO;
        }
      }
    }
    block++;
  }
  (*mute) = 0;

  return NONE_MUTE;
}

static int write_drum(const unsigned char data, FILE *ofp)
{
  int status;

  if (data < 35) { 
    status = fprintf(ofp, "       ;%s\n", DRUM_COMMENT[47]);
  }
  else if (data > 81) {
    status = fprintf(ofp, "       ;%s\n", DRUM_COMMENT[47]);
  }
  else {
    status = fprintf(ofp, "       ;%s\n", DRUM_COMMENT[data - 35]);
  }

  if (status < 0) { return E_W_WRITE_DATA; }

  return STATUS_SUCCESS;
}

static int write_ccom(const unsigned char ctrl, FILE *ofp)
{
  int status;

  if (ctrl > 127) {
    status = fputs("\n", ofp);
    if (status == EOF) { return E_W_WRITE_DATA; }
  }
  else {
    status = fprintf(ofp, "                   ;%s\n", CTRL_COMMENT[ctrl]);
    if (status < 0) { return E_W_WRITE_DATA; }
  }


  return STATUS_SUCCESS;
}

static int write_meta(const unsigned char *event, FILE *ofp)
{
  int type;
  int len;
  int cnt;
  int i;
  int status;

  type = (int) event[1];
  status = fprintf(ofp, "  $%02X", (unsigned int) type);
  if (status < 0) { return E_W_WRITE_DATA; }
  event += 2;

  cnt = check_len(event, &len);
  if (cnt <= 0) { return cnt; }

  status = fputs("  $", ofp);
  if (status == EOF) { return E_W_WRITE_DATA; }

  while (cnt) {
    status = fprintf(ofp, "%02X", (unsigned int) *event);
    if (status < 0) { return E_W_WRITE_DATA; }
    event++;
    cnt--;
  }

  if (len) {
    status = fputs("  $", ofp);
    if (status == EOF) { return E_W_WRITE_DATA; }
    for (i = 0; i < len; i++) {
      status = fprintf(ofp, "%02X", (unsigned int) event[i]);
      if (status < 0) { return E_W_WRITE_DATA; }
    }
  }

  status = write_mcom(event, type, len, ofp);
  if (status < 0) { return status; }

  return STATUS_SUCCESS;
}

static int write_syse(const unsigned char *event, FILE *ofp)
{
  int cnt;
  int len;
  int i;
  int status;

  event++;

  cnt = check_len(event, &len);
  if (cnt <= 0) { return cnt; }

  status = fputs("  $", ofp);
  if (status == EOF) { return E_W_WRITE_DATA; }

  while (cnt) {
    status = fprintf(ofp, "%02X", (unsigned int) *event);
    if (status < 0) { return E_W_WRITE_DATA; }
    event++;
    cnt--;
  }

  if (len) {
    status = fputs("  $", ofp);
    if (status == EOF) { return E_W_WRITE_DATA; }
    for (i = 0; i<len; i++) {
      status = fprintf(ofp, "%02X", (unsigned int) event[i]);
      if (status < 0) { return E_W_WRITE_DATA; }
    }
  }
  status = write_spad(len, ofp);
  if (status < 0) { return E_W_WRITE_DATA; }

  status = fputs(" ;System Exclusive\n", ofp);
  if (status == EOF) { return E_W_WRITE_DATA; }

  return STATUS_SUCCESS;
}

static int check_len(const unsigned char *event, int *value)
{
  int data;

  *value = *event;
  if ((*value & 0x80) == 0) { return 1; }

  *value = (*value & 0x7F) << 7;
  data = (int) event[1];
  if ((data & 0x80) == 0) { *value |= data; return 2; }

  *value = (*value | (data & 0x7F)) << 7;
  data = (int) event[2];
  if ((data & 0x80) == 0) { *value |= data; return 3; }

  *value = (*value | (data & 0x7F)) << 7;
  data = (int) event[3];
  if ((data & 0x80) == 0) { *value |= data; return 4; }

  return E_W_LEN_FORMAT;
}

static int write_mcom(const unsigned char *data, int type, int len, FILE *ofp)
{
  int status;

  switch (type) {
  case 0x00:
    if (len != 2) {
      status = fputs("\n", ofp);
      if (status == EOF) { return E_W_WRITE_COM; }
      return STATUS_SUCCESS;
    }
    status = fputs("            ;Sequence Number\n", ofp);
    if (status == EOF) { return E_W_WRITE_COM; }
    break;
  case 0x01:
    status = write_mpad(len, ofp);
    if (status < 0) { return status; }
    status = fputs(" ;Text Event\n", ofp);
    if (status == EOF) { return E_W_WRITE_COM; }
    status = write_text(data, len, ofp);
    if (status < 0) { return status; }
    break;
  case 0x02:
    status = write_mpad(len, ofp);
    if (status < 0) { return status; }
    status = fputs(" ;Copyright Notice\n", ofp);
    if (status == EOF) { return E_W_WRITE_COM; }
    status = write_text(data, len, ofp);
    if (status < 0) { return status; }
    break;
  case 0x03:
    status = write_mpad(len, ofp);
    if (status < 0) { return status; }
    status = fputs(" ;Sequence/Track Name\n", ofp);
    if (status == EOF) { return E_W_WRITE_COM; }
    status = write_text(data, len, ofp);
    if (status < 0) { return status; }
    break;
  case 0x04:
    status = write_mpad(len, ofp);
    if (status < 0) { return status; }
    status = fputs(" ;Instrument Name\n", ofp);
    if (status == EOF) { return E_W_WRITE_COM; }
    status = write_text(data, len, ofp);
    if (status < 0) { return status; }
    break;
  case 0x05:
    status = write_mpad(len, ofp);
    if (status < 0) { return status; }
    status = fputs(" ;Lyric\n", ofp);
    if (status == EOF) { return E_W_WRITE_COM; }
    status = write_text(data, len, ofp);
    if (status < 0) { return status; }
    break;
  case 0x06:
    status = write_mpad(len, ofp);
    if (status < 0) { return status; }
    status = fputs(" ;Marker\n", ofp);
    if (status == EOF) { return E_W_WRITE_COM; }
    status = write_text(data, len, ofp);
    if (status < 0) { return status; }
    break;
  case 0x07:
    status = write_mpad(len, ofp);
    if (status < 0) { return status; }
    status = fputs(" ;Cue Point\n", ofp);
    if (status == EOF) { return E_W_WRITE_COM; }
    status = write_text(data, len, ofp);
    if (status < 0) { return status; }
    break;
  case 0x08: case 0x09: case 0x0A: case 0x0B:
  case 0x0C: case 0x0D: case 0x0E: case 0x0F:
    status = fputs("\n", ofp);
    if (status == EOF) { return E_W_WRITE_COM; }
    status = write_text(data, len, ofp);
    if (status < 0) { return E_W_WRITE_COM; }
    break;
  case 0x20:
    if (len != 1) {
      status = fputs("\n", ofp);
      if (status == EOF) { return E_W_WRITE_COM; }
      return STATUS_SUCCESS;
    }
    status = fputs("              ;MIDI Channel Prefix\n", ofp);
    if (status == EOF) { return E_W_WRITE_COM; }
    break;
  case 0x2F:
    if (len != 0) {
      status = fputs("\n", ofp);
      if (status == EOF) { return E_W_WRITE_COM; }
      return STATUS_SUCCESS;
    }
    status = fputs("                   ;End of Track\n", ofp);
    if (status == EOF) { return E_W_WRITE_COM; }
    break;
  case 0x51:
    if (len != 3) {
      status = fputs("\n", ofp);
      if (status == EOF) { return E_W_WRITE_COM; }
      return STATUS_SUCCESS;
    }
    status = fputs("          ;Set Tempo", ofp);
    if (status == EOF) { return E_W_WRITE_COM; }
    status = write_tempo(data, ofp);
    if (status < 0) { return status; }
    break;
  case 0x54:
    if (len != 5) {
      status = fputs("\n", ofp);
      if (status == EOF) { return E_W_WRITE_COM; }
      return STATUS_SUCCESS;
    }
    status = fputs("      ;SMTP Offset\n", ofp);
    if (status == EOF) { return E_W_WRITE_COM; }
    break;
  case 0x58:
    if (len != 4) {
      status = fputs("\n", ofp);
      if (status == EOF) { return E_W_WRITE_COM; }
      return STATUS_SUCCESS;
    }
    status = fputs("        ;Time Signature", ofp);
    if (status == EOF) { return E_W_WRITE_COM; }
    status = write_tsig(data, ofp);
    if (status < 0) { return status; }
    break;
  case 0x59:
    if (len != 2) {
      status = fputs("\n", ofp);
      if (status == EOF) { return E_W_WRITE_COM; }
      return STATUS_SUCCESS;
    }
    status = fputs("            ;Key Signatuer", ofp);
    if (status == EOF) { return E_W_WRITE_COM; }
    status = write_ksig(data, ofp);
    if (status < 0) { return status; }
    break;
  case 0x7F:
    status = write_mpad(len, ofp);
    if (status < 0) { return status; }
    status = fputs(" ;Sequencer Spcific\n", ofp);
    if (status == EOF) { return E_W_WRITE_COM; }
    break;
  default:
    status = fputs("\n", ofp);
    if (status == EOF) { return E_W_WRITE_COM; }
    break;
  }

  return STATUS_SUCCESS;
}

static int write_mpad(const int len, FILE *ofp)
{
  int status;
  switch (len) {
  case 0: status = fputs("                  ", ofp); break;
  case 1: status = fputs("             ", ofp); break;
  case 2: status = fputs("           ", ofp); break;
  case 3: status = fputs("         ", ofp); break;
  case 4: status = fputs("       ", ofp); break;
  case 5: status = fputs("     ", ofp); break;
  case 6: status = fputs("   ", ofp); break;
  case 7: status = fputs(" ", ofp); break;
  default: return STATUS_SUCCESS;
  }
  if (status == EOF) { return E_W_WRITE_COM; }

  return STATUS_SUCCESS;
}

static int write_text(const unsigned char *data, int len, FILE *ofp)
{
  int i;
  int status;

  if (len == 0) { return STATUS_SUCCESS; }

  status = fputs("                         ;\"", ofp);
  if (status == EOF) { return E_W_WRITE_TEXT; }

  for (i = 0; i < len; i++) {
    if (data[i] == '\n') {
      status = fputs("\\n", ofp);
      if (status == EOF) { return E_W_WRITE_TEXT; }
    }
    else if (data[i] == '\r') {
      status = fputs("\\r", ofp);
      if (status == EOF) { return E_W_WRITE_TEXT; }
    }
    else if (data[i] == '\"') {
      status = fputs("\\\"", ofp);
      if (status == EOF) { return E_W_WRITE_TEXT; }
    }
    else {
      status = fprintf(ofp, "%c", data[i]);
      if (status < 0) { return E_W_WRITE_TEXT; }
    }
  }
  status = fputs("\"\n", ofp);
  if (status == EOF) { return E_W_WRITE_TEXT; }
  return STATUS_SUCCESS;
}

static int write_tempo(const unsigned char *data, FILE *ofp)
{
  int bpm;
  int status;

  bpm = ((int) data[0] << 16) + ((int) data[1] << 8) + (int) data[2];
  if (bpm == 0) { return E_W_WRITE_TEMPO; }
  bpm = (60 * 1000000) / bpm;
  status = fprintf(ofp, " (@%d)\n", bpm);
  if (status < 0) { return E_W_WRITE_COM; }

  return STATUS_SUCCESS;
}

static int write_tsig(const unsigned char *data, FILE *ofp)
{
  int status;

  if (data[1] < 32) {
    status = fprintf(ofp, " (%d/%u)\n", (int) data[0], 1U << data[1]);
    if (status < 0) { return E_W_WRITE_TIMESIG; }
    return STATUS_SUCCESS;
  }

  status = fprintf(ofp, " (%d/", (int) data[0]);
  if (status < 0) { return E_W_WRITE_TIMESIG; }

  status = write_exp((int) data[1], ofp);
  if (status < 0) { return status; }

  status = fputs(")\n", ofp);
  if (status == EOF) { return E_W_WRITE_TIMESIG; }

  return STATUS_SUCCESS;
}

static int write_exp(int cnt, FILE *ofp)
{
  unsigned char exp[39];
  int len = 1;
  int i;
  int crr;
  int status;

  for (i = 1; i < 39; i++) { exp[i] = 0; }
  exp[0] = 1;
  if (cnt & 0x80) { return E_W_TIMESIG_VALUE; }

  for ( ; cnt; cnt--) {
    crr = 0;
    for (i = 0; i < len; i++) {
      exp[i] = (unsigned char) ((exp[i] << 1) + crr);
      if (exp[i] > 9) {
        crr = 1;
        exp[i] = (unsigned char) (exp[i] - 10);
      }
      else { crr = 0; }
    }
    if (crr == 1) {
      exp[i] = 1;
      len++;
    }
  }

  for ( ; len; len--) {
    status = fprintf(ofp, "%01d", (int) exp[len - 1]);
    if (status < 0) { return E_W_WRITE_TIMESIG; }
  }
  return STATUS_SUCCESS;
}

static int write_ksig(const unsigned char *data, FILE *ofp)
{
  int status;
  const char * const sharp[] = {
    "C",  "G",  "D",  "A",  "E",  "B",  "F#", "C#", "G#", "D#", "A#", "F"
  };
  const char * const flat[] = {
    "F",  "C",  "G",  "D",  "A",  "E",  "B",  "Gb", "Db", "Ab", "Eb", "Bb"
  };

  if (data[0] & 0x80) {
    if (data[1]) {
      status = fprintf(ofp, " (%sm)\n", flat[(data[0] % 12)]);
    }
    else {
      status = fprintf(ofp, " (%s)\n", flat[((data[0] - 3) % 12)]);
    }
  }
  else {
    if (data[1]) {
      status = fprintf(ofp, " (%sm)\n", sharp[((data[0] + 3) % 12)]);
    }
    else {
      status = fprintf(ofp, " (%s)\n", sharp[(data[0] % 12)]);
    }
  }

  if (status < 0) { return E_W_WRITE_KEY; }

  return STATUS_SUCCESS;
}

static int write_spad(const int len, FILE *ofp)
{
  int status;
  switch (len) {
  case 0: status = fputs("                       ", ofp); break;
  case 1: status = fputs("                  ", ofp); break;
  case 2: status = fputs("                ", ofp); break;
  case 3: status = fputs("              ", ofp); break;
  case 4: status = fputs("            ", ofp); break;
  case 5: status = fputs("          ", ofp); break;
  case 6: status = fputs("        ", ofp); break;
  case 7: status = fputs("      ", ofp); break;
  case 8: status = fputs("    ", ofp); break;
  case 9: status = fputs("  ", ofp); break;
  default: return STATUS_SUCCESS;
  }

  if (status == EOF) { return E_W_WRITE_COM; }

  return STATUS_SUCCESS;
}

static int write_dcom(const unsigned char prog, FILE *ofp)
{
  int status;

  switch (prog) {
  case   0:  status = fputs(";Standard Set\n", ofp); break;
  case   8:  status = fputs(";Room Set\n", ofp); break;
  case  16:  status = fputs(";Power Set\n", ofp); break;
  case  24:  status = fputs(";Electronic Set\n", ofp); break;
  case  25:  status = fputs(";Analog Set\n", ofp); break;
  case  32:  status = fputs(";Jazz Set\n", ofp); break;
  case  40:  status = fputs(";Brush Set\n", ofp); break;
  case  48:  status = fputs(";Orchestra Set\n", ofp); break;
  case  56:  status = fputs(";SFX Set\n", ofp); break;
  default:   status = fputs(";Undefined Drum Set\n", ofp); break;
  }
  if (status == EOF) { return E_W_WRITE_COM; }

  return STATUS_SUCCESS;
}

static int write_pcom(const unsigned char prog, FILE *ofp)
{
  int status;

  if (prog > 127) {
    status = fputs(";Undefined Instrument\n", ofp);
    if (status == EOF) { return E_W_WRITE_DATA; }
  }
  else {
    status = fprintf(ofp, ";%s\n", PROG_COMMENT[prog]);
    if (status < 0) { return E_W_WRITE_DATA; }
  }

  return STATUS_SUCCESS;
}

static int make_bar(const unsigned char *event, struct wmbt_st *inf)
{
  struct wbar_st *bar;

  if (event[0] != 0xFF) { return STATUS_SUCCESS; }

  if (event[1] == 0x59) {
    if (event[3] & 0x80) {
      inf->bar[inf->mbt[0]].key = NOTE_OFFSET;
    }
    else {
      inf->bar[inf->mbt[0]].key = 0;
    }
    return STATUS_SUCCESS;
  }

  if (inf->trk != 0) { return STATUS_SUCCESS; }

  if (inf->mbt[1] != 1) { return STATUS_SUCCESS; }

  if (inf->mbt[2] != 0) { return STATUS_SUCCESS; }

  if (event[1] == 0x58) {
    bar = &(inf->bar[inf->mbt[0]]);

    bar->beat.num = (int) event[3];
    if (bar->beat.num <= 0) { return E_W_BEAT_DATA; }

    if (event[4] >= 2) {
      bar->beat.den = inf->tpqn >> (event[4] - 2);
    }
    else {
      bar->beat.den = inf->tpqn << (2 - event[4]);
    }

    if (bar->beat.den <= 0) { return E_W_BEAT_DATA; }

    bar[1].tick = bar->tick + bar->beat.num * bar->beat.den;
  }

  return STATUS_SUCCESS;
}
